from __future__ import annotations

import asyncio
import functools
import inspect
import sys
import textwrap
from abc import ABCMeta
from concurrent.futures import CancelledError
from dataclasses import dataclass, field
from itertools import chain
from types import TracebackType
from typing import Any, cast

from wrapt import ObjectProxy  # type: ignore

from coredis._utils import b, hash_slot
from coredis.client import Client, Redis, RedisCluster
from coredis.commands._key_spec import KeySpec
from coredis.commands.constants import CommandName, NodeFlag
from coredis.commands.script import Script
from coredis.connection import BaseConnection, ClusterConnection, CommandInvocation
from coredis.exceptions import (
    AskError,
    ClusterCrossSlotError,
    ClusterDownError,
    ClusterTransactionError,
    ConnectionError,
    ExecAbortError,
    MovedError,
    RedisClusterException,
    RedisError,
    ResponseError,
    TimeoutError,
    TryAgainError,
    WatchError,
)
from coredis.pool import ClusterConnectionPool, ConnectionPool
from coredis.pool.nodemanager import ManagedNode
from coredis.response._callbacks import (
    AnyStrCallback,
    AsyncPreProcessingCallback,
    BoolCallback,
    BoolsCallback,
    NoopCallback,
    SimpleStringCallback,
)
from coredis.retry import ConstantRetryPolicy, retryable
from coredis.typing import (
    AnyStr,
    Callable,
    Coroutine,
    Dict,
    Generic,
    Iterable,
    KeyT,
    List,
    Optional,
    Parameters,
    ParamSpec,
    ResponseType,
    Set,
    StringT,
    Tuple,
    Type,
    TypeVar,
    ValueT,
)

P = ParamSpec("P")
R = TypeVar("R")
T = TypeVar("T")

ERRORS_ALLOW_RETRY = (
    MovedError,
    AskError,
    TryAgainError,
)

UNWATCH_COMMANDS = {CommandName.DISCARD, CommandName.EXEC, CommandName.UNWATCH}


def wrap_pipeline_method(
    kls: PipelineMeta, func: Callable[P, Coroutine[Any, Any, R]]
) -> Callable[P, Coroutine[Any, Any, R]]:
    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        return await func(*args, **kwargs)

    wrapper.__annotations__ = wrapper.__annotations__.copy()
    wrapper.__annotations__["return"] = kls
    wrapper.__doc__ = textwrap.dedent(wrapper.__doc__ or "")
    wrapper.__doc__ = f"""
Pipeline variant of :meth:`coredis.Redis.{func.__name__}` that does not execute
immediately and instead pushes the command into a stack for batch send
and returns the instance of :class:`{kls.__name__}` itself.

To fetch the return values call :meth:`{kls.__name__}.execute` to process the pipeline
and retrieve responses for the commands executed in the pipeline.

{wrapper.__doc__}
"""
    return wrapper


@dataclass
class PipelineCommand:
    command: bytes
    args: Tuple[ValueT, ...]
    callback: Callable[..., Any] = NoopCallback()  # type: ignore
    options: Dict[str, Optional[ValueT]] = field(default_factory=dict)
    request: Optional[asyncio.Future[ResponseType]] = None


@dataclass
class ClusterPipelineCommand(PipelineCommand):
    position: int = 0
    result: Optional[Any] = None  # type: ignore
    asking: bool = False


class NodeCommands:
    def __init__(
        self,
        client: RedisCluster[AnyStr],
        connection: ClusterConnection,
        in_transaction: bool = False,
        timeout: Optional[float] = None,
    ):
        self.client = client
        self.connection = connection
        self.commands: List[ClusterPipelineCommand] = []
        self.in_transaction = in_transaction
        self.timeout = timeout

    def extend(self, c: List[ClusterPipelineCommand]) -> None:
        self.commands.extend(c)

    def append(self, c: ClusterPipelineCommand) -> None:
        self.commands.append(c)

    async def write(self) -> None:
        connection = self.connection
        commands = self.commands

        # We are going to clobber the commands with the write, so go ahead
        # and ensure that nothing is sitting there from a previous run.

        for c in commands:
            c.result = None

        # build up all commands into a single request to increase network perf
        # send all the commands and catch connection and timeout errors.
        try:
            requests = await connection.create_requests(
                [
                    CommandInvocation(
                        cmd.command,
                        cmd.args,
                        (
                            bool(cmd.options.get("decode"))
                            if cmd.options.get("decode")
                            else None
                        ),
                        None,
                    )
                    for cmd in commands
                ],
                timeout=self.timeout,
            )
            for i, cmd in enumerate(commands):
                cmd.request = requests[i]
        except (ConnectionError, TimeoutError) as e:
            for c in commands:
                c.result = e

    async def read(self) -> None:
        connection = self.connection
        success = True

        for c in self.commands:
            if c.result is None:
                try:
                    c.result = await c.request if c.request else None
                except ExecAbortError:
                    raise
                except (ConnectionError, TimeoutError, RedisError) as e:
                    success = False
                    c.result = e

        if self.in_transaction:
            transaction_result = []
            if success:
                for c in self.commands:
                    if c.command == CommandName.EXEC:
                        if c.result:
                            transaction_result = cast(List[ResponseType], c.result)
                        else:
                            raise WatchError("Watched variable changed.")
                for idx, c in enumerate(
                    [
                        _c
                        for _c in sorted(self.commands, key=lambda x: x.position)
                        if _c.command not in {CommandName.MULTI, CommandName.EXEC}
                    ]
                ):
                    if isinstance(c.callback, AsyncPreProcessingCallback):
                        await c.callback.pre_process(
                            self.client, transaction_result[idx], **c.options
                        )
                    c.result = c.callback(
                        transaction_result[idx],
                        version=connection.protocol_version,
                        **c.options,
                    )
            elif isinstance(self.commands[0].result, BaseException):
                raise self.commands[0].result


class PipelineMeta(ABCMeta):
    RESULT_CALLBACKS: Dict[str, Callable[..., Any]]
    NODES_FLAGS: Dict[str, NodeFlag]

    def __new__(cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, object]):
        kls = super().__new__(cls, name, bases, namespace)

        for name, method in PipelineMeta.get_methods(kls).items():
            if getattr(method, "__coredis_command", None):
                setattr(kls, name, wrap_pipeline_method(kls, method))

        return kls

    @staticmethod
    def get_methods(kls: PipelineMeta) -> Dict[str, Callable[..., Any]]:
        return dict(k for k in inspect.getmembers(kls) if inspect.isfunction(k[1]))


class ClusterPipelineMeta(PipelineMeta):
    def __new__(cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, object]):
        kls = super().__new__(cls, name, bases, namespace)
        for name, method in ClusterPipelineMeta.get_methods(kls).items():
            cmd = getattr(method, "__coredis_command", None)
            if cmd:
                if cmd.cluster.route:
                    kls.NODES_FLAGS[cmd.command] = cmd.cluster.route
                if cmd.cluster.multi_node:
                    kls.RESULT_CALLBACKS[cmd.command] = cmd.cluster.combine or (
                        lambda r, **_: r
                    )
                else:
                    kls.RESULT_CALLBACKS[cmd.command] = lambda response, **_: list(
                        response.values()
                    ).pop()
        return kls


class PipelineImpl(Client[AnyStr], metaclass=PipelineMeta):
    """Pipeline for the Redis class"""

    """
    Pipelines provide a way to transmit multiple commands to the Redis server
    in one transmission.  This is convenient for batch processing, such as
    saving all the values in a list to Redis.

    All commands executed within a pipeline are wrapped with MULTI and EXEC
    calls. This guarantees all commands executed in the pipeline will be
    executed atomically.

    Any command raising an exception does *not* halt the execution of
    subsequent commands in the pipeline. Instead, the exception is caught
    and its instance is placed into the response list returned by execute().
    Code iterating over the response list should be able to deal with an
    instance of an exception as a potential value. In general, these will be
    ResponseError exceptions, such as those raised when issuing a command
    on a key of a different datatype.
    """

    command_stack: List[PipelineCommand]
    connection_pool: ConnectionPool

    def __init__(
        self,
        client: Client[AnyStr],
        transaction: Optional[bool],
        watches: Optional[Parameters[KeyT]] = None,
        timeout: Optional[float] = None,
    ) -> None:
        self.client = client
        self.connection_pool = client.connection_pool
        self.connection = None
        self._transaction = transaction
        self.watching = False
        self.watches: Optional[Parameters[KeyT]] = watches or None
        self.command_stack = []
        self.cache = None  # not implemented.
        self.explicit_transaction = False
        self.scripts: Set[Script[AnyStr]] = set()
        self.timeout = timeout

    async def __aenter__(self) -> "PipelineImpl[AnyStr]":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.reset_pipeline()

    def __len__(self) -> int:
        return len(self.command_stack)

    def __bool__(self) -> bool:
        return True

    async def reset_pipeline(self) -> None:
        self.command_stack.clear()
        self.scripts: Set[Script[AnyStr]] = set()
        # make sure to reset the connection state in the event that we were
        # watching something

        if self.watching and self.connection:
            try:
                # call this manually since our unwatch or
                # immediate_execute_command methods can call reset_pipeline()
                request = await self.connection.create_request(
                    CommandName.UNWATCH, decode=False
                )
                await request
            except ConnectionError:
                # disconnect will also remove any previous WATCHes
                self.connection.disconnect()
        # clean up the other instance attributes
        self.watching = False
        self.watches = []
        self.explicit_transaction = False
        # we can safely return the connection to the pool here since we're
        # sure we're no longer WATCHing anything

        if self.connection:
            self.connection_pool.release(self.connection)
            self.connection = None

    def multi(self) -> None:
        """
        Starts a transactional block of the pipeline after WATCH commands
        are issued. End the transactional block with `execute`.
        """

        if self.explicit_transaction:
            raise RedisError("Cannot issue nested calls to MULTI")

        if self.command_stack:
            raise RedisError(
                "Commands without an initial WATCH have already been issued"
            )
        self.explicit_transaction = True

    async def execute_command(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., Any] = NoopCallback(),  # type: ignore
        **options: Optional[ValueT],
    ) -> PipelineImpl[AnyStr]:  # type: ignore
        if (
            self.watching or command == CommandName.WATCH
        ) and not self.explicit_transaction:
            return await self.immediate_execute_command(
                command, *args, callback=callback, **options
            )  # type: ignore

        return self.pipeline_execute_command(
            command, *args, callback=callback, **options
        )

    async def immediate_execute_command(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., Any] = NoopCallback(),  # type: ignore
        **kwargs: Optional[ValueT],
    ) -> Any:  # type: ignore
        """
        Executes a command immediately, but don't auto-retry on a
        ConnectionError if we're already WATCHing a variable. Used when
        issuing WATCH or subsequent commands retrieving their values but before
        MULTI is called.

        :meta private:
        """
        conn = self.connection
        # if this is the first call, we need a connection

        if not conn:
            conn = await self.connection_pool.get_connection()
            self.connection = conn
        try:
            request = await conn.create_request(
                command, *args, decode=kwargs.get("decode")
            )

            return callback(
                await request,
                version=conn.protocol_version,
                **kwargs,
            )
        except (ConnectionError, TimeoutError):
            conn.disconnect()

            # if we're not already watching, we can safely retry the command
            try:
                if not self.watching:
                    request = await conn.create_request(
                        command, *args, decode=kwargs.get("decode")
                    )
                    return callback(
                        await request, version=conn.protocol_version, **kwargs
                    )
            except ConnectionError:
                # the retry failed so cleanup.
                conn.disconnect()
                await self.reset_pipeline()
                raise
        finally:
            if command in UNWATCH_COMMANDS:
                self.watching = False
            elif command == CommandName.WATCH:
                self.watching = True

    def pipeline_execute_command(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., Any],
        **options: Optional[ValueT],
    ) -> PipelineImpl[AnyStr]:
        """
        Stages a command to be executed next execute() invocation

        Returns the current Pipeline object back so commands can be
        chained together, such as:

        pipe = pipe.set('foo', 'bar').incr('baz').decr('bang')

        At some other point, you can then run: pipe.execute(),
        which will execute all commands queued in the pipe.

        :meta private:
        """
        self.command_stack.append(
            PipelineCommand(
                command=command, args=args, options=options, callback=callback
            )
        )

        return self

    async def _execute_transaction(
        self,
        connection: BaseConnection,
        commands: List[PipelineCommand],
        raise_on_error: bool,
    ) -> Tuple[Any, ...]:
        cmds = list(
            chain(
                [
                    PipelineCommand(
                        command=CommandName.MULTI,
                        args=(),
                    )
                ],
                commands,
                [
                    PipelineCommand(
                        command=CommandName.EXEC,
                        args=(),
                    )
                ],
            )
        )
        if self.watches:
            await self.watch(*self.watches)

        requests = await connection.create_requests(
            [
                CommandInvocation(
                    cmd.command,
                    cmd.args,
                    (
                        bool(cmd.options.get("decode"))
                        if cmd.options.get("decode")
                        else None
                    ),
                    None,
                )
                for cmd in cmds
            ],
            timeout=self.timeout,
        )
        for i, cmd in enumerate(cmds):
            cmd.request = requests[i]

        errors: List[Tuple[int, Optional[RedisError]]] = []
        multi_failed = False

        # parse off the response for MULTI
        # NOTE: we need to handle ResponseErrors here and continue
        # so that we read all the additional command messages from
        # the socket
        try:
            await cmds[0].request if cmds[0].request else None
        except RedisError:
            multi_failed = True
            errors.append((0, cast(RedisError, sys.exc_info()[1])))

        # and all the other commands
        for i, cmd in enumerate(cmds[1:-1]):
            try:
                if cmd.request:
                    assert (await cmd.request) in {b"QUEUED", "QUEUED"}
            except RedisError:
                ex = cast(RedisError, sys.exc_info()[1])
                self.annotate_exception(ex, i + 1, cmd.command, cmd.args)
                errors.append((i, ex))

        response: List[ResponseType]
        try:
            response = cast(
                List[ResponseType],
                await cmds[-1].request if cmds[-1].request else None,
            )
        except (ExecAbortError, ResponseError):
            if self.explicit_transaction and not multi_failed:
                await self.immediate_execute_command(
                    CommandName.DISCARD, callback=BoolCallback()
                )

            if errors and errors[0][1]:
                raise errors[0][1]
            raise

        if response is None:
            raise WatchError("Watched variable changed.")

        # put any parse errors into the response

        for i, e in errors:
            response.insert(i, cast(ResponseType, e))

        if len(response) != len(commands):
            if self.connection:
                self.connection.disconnect()
            raise ResponseError(
                "Wrong number of response items from pipeline execution"
            )

        # find any errors in the response and raise if necessary
        if raise_on_error:
            self.raise_first_error(commands, response)

        # We have to run response callbacks manually
        data: List[Any] = []
        for r, cmd in zip(response, commands):
            if not isinstance(r, Exception):
                if isinstance(cmd.callback, AsyncPreProcessingCallback):
                    await cmd.callback.pre_process(self.client, r, **cmd.options)
                r = cmd.callback(r, version=connection.protocol_version, **cmd.options)
            data.append(r)
        return tuple(data)

    async def _execute_pipeline(
        self,
        connection: BaseConnection,
        commands: List[PipelineCommand],
        raise_on_error: bool,
    ) -> Tuple[Any, ...]:
        # build up all commands into a single request to increase network perf
        requests = await connection.create_requests(
            [
                CommandInvocation(
                    cmd.command,
                    cmd.args,
                    (
                        bool(cmd.options.get("decode"))
                        if cmd.options.get("decode")
                        else None
                    ),
                    None,
                )
                for cmd in commands
            ],
            timeout=self.timeout,
        )
        for i, cmd in enumerate(commands):
            cmd.request = requests[i]

        response: List[Any] = []

        for cmd in commands:
            try:
                res = await cmd.request if cmd.request else None
                if isinstance(cmd.callback, AsyncPreProcessingCallback):
                    await cmd.callback.pre_process(self.client, res, **cmd.options)
                response.append(
                    cmd.callback(
                        res,
                        version=connection.protocol_version,
                        **cmd.options,
                    )
                )
            except ResponseError:
                response.append(sys.exc_info()[1])

        if raise_on_error:
            self.raise_first_error(commands, response)

        return tuple(response)

    def raise_first_error(
        self, commands: List[PipelineCommand], response: ResponseType
    ) -> None:
        assert isinstance(response, list)
        for i, r in enumerate(response):
            if isinstance(r, RedisError):
                self.annotate_exception(r, i + 1, commands[i].command, commands[i].args)
                raise r

    def annotate_exception(
        self,
        exception: Optional[RedisError],
        number: int,
        command: bytes,
        args: Iterable[ValueT],
    ) -> None:
        if exception:
            cmd = command.decode("latin-1")
            args = " ".join(map(str, args))
            msg = "Command # {} ({} {}) of pipeline caused error: {}".format(
                number,
                cmd,
                args,
                str(exception.args[0]),
            )
            exception.args = (msg,) + exception.args[1:]

    async def load_scripts(self):
        # make sure all scripts that are about to be run on this pipeline exist
        scripts = list(self.scripts)
        immediate = self.immediate_execute_command
        shas = [s.sha for s in scripts]
        # we can't use the normal script_* methods because they would just
        # get buffered in the pipeline.
        exists = await immediate(
            CommandName.SCRIPT_EXISTS, *shas, callback=BoolsCallback()
        )

        if not all(exists):
            for s, exist in zip(scripts, exists):
                if not exist:
                    s.sha = await immediate(
                        CommandName.SCRIPT_LOAD,
                        s.script,
                        callback=AnyStrCallback[AnyStr](),
                    )

    async def execute(self, raise_on_error: bool = True) -> Tuple[Any, ...]:
        """Executes all the commands in the current pipeline"""
        stack = self.command_stack

        if not stack:
            return ()

        if self.scripts:
            await self.load_scripts()

        if self._transaction or self.explicit_transaction:
            exec = self._execute_transaction
        else:
            exec = self._execute_pipeline

        conn = self.connection

        if not conn:
            conn = await self.connection_pool.get_connection()
            # assign to self.connection so reset_pipeline() releases the connection
            # back to the pool after we're done
            self.connection = conn

        try:
            return await exec(conn, stack, raise_on_error)
        except (ConnectionError, TimeoutError, CancelledError):
            conn.disconnect()

            # if we were watching a variable, the watch is no longer valid
            # since this connection has died. raise a WatchError, which
            # indicates the user should retry his transaction. If this is more
            # than a temporary failure, the WATCH that the user next issues
            # will fail, propegating the real ConnectionError

            if self.watching:
                raise WatchError(
                    "A ConnectionError occured on while watching one or more keys"
                )
            # otherwise, it's safe to retry since the transaction isn't
            # predicated on any state

            return await exec(conn, stack, raise_on_error)
        finally:
            await self.reset_pipeline()

    async def watch(self, *keys: KeyT) -> bool:
        """
        Watches the values at keys ``keys``
        """

        if self.explicit_transaction:
            raise RedisError("Cannot issue a WATCH after a MULTI")

        return await self.immediate_execute_command(
            CommandName.WATCH, *keys, callback=SimpleStringCallback()
        )

    async def unwatch(self) -> bool:
        """Unwatches all previously specified keys"""

        return (
            await self.immediate_execute_command(
                CommandName.UNWATCH, callback=SimpleStringCallback()
            )
            if self.watching
            else True
        )


class ClusterPipelineImpl(Client[AnyStr], metaclass=ClusterPipelineMeta):
    client: RedisCluster[AnyStr]
    connection_pool: ClusterConnectionPool
    command_stack: List[ClusterPipelineCommand]

    RESULT_CALLBACKS: Dict[str, Callable[..., Any]] = {}
    NODES_FLAGS: Dict[str, NodeFlag] = {}

    def __init__(
        self,
        client: RedisCluster[AnyStr],
        transaction: Optional[bool] = False,
        watches: Optional[Parameters[KeyT]] = None,
        timeout: Optional[float] = None,
    ) -> None:
        self.command_stack = []
        self.refresh_table_asap = False
        self.client = client
        self.connection_pool = client.connection_pool
        self.result_callbacks = client.result_callbacks
        self._transaction = transaction
        self._watched_node: Optional[ManagedNode] = None
        self._watched_connection: Optional[ClusterConnection] = None
        self.watches: Optional[Parameters[KeyT]] = watches or None
        self.watching = False
        self.explicit_transaction = False
        self.cache = None  # not implemented.
        self.timeout = timeout

    async def watch(self, *keys: KeyT) -> bool:
        if self.explicit_transaction:
            raise RedisError("Cannot issue a WATCH after a MULTI")

        return await self.immediate_execute_command(
            CommandName.WATCH, *keys, callback=SimpleStringCallback()
        )

    async def unwatch(self) -> bool:
        if self._watched_connection:
            try:
                return await self._unwatch(self._watched_connection)
            finally:
                if self._watched_connection:
                    self.connection_pool.release(self._watched_connection)
                    self.watching = False
                    self._watched_node = None
                    self._watched_connection = None
        return True

    def __repr__(self):
        return f"{type(self).__name__}"

    def __del__(self):
        if self._watched_connection:
            self.connection_pool.release(self._watched_connection)

    def __len__(self):
        return len(self.command_stack)

    def __bool__(self) -> bool:
        return True

    async def __aenter__(self) -> "ClusterPipelineImpl[AnyStr]":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.reset_pipeline()

    async def execute_command(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., Any] = NoopCallback(),  # type: ignore
        **options: Optional[ValueT],
    ) -> ClusterPipelineImpl[AnyStr]:  # type: ignore
        if (
            self.watching or command == CommandName.WATCH
        ) and not self.explicit_transaction:
            return await self.immediate_execute_command(
                command, *args, callback=callback, **options
            )  # type: ignore
        return self.pipeline_execute_command(
            command, *args, callback=callback, **options
        )

    def pipeline_execute_command(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., Any],
        **options: Optional[ValueT],
    ) -> ClusterPipelineImpl[AnyStr]:
        self.command_stack.append(
            ClusterPipelineCommand(
                command=command,
                args=args,
                options=options,
                callback=callback,
                position=len(self.command_stack),
            )
        )

        return self

    def raise_first_error(self) -> None:
        for c in self.command_stack:
            r = c.result

            if isinstance(r, RedisError):
                self.annotate_exception(r, c.position + 1, c.command, c.args)
                raise r

    def annotate_exception(
        self,
        exception: Optional[RedisError],
        number: int,
        command: bytes,
        args: Iterable[ValueT],
    ) -> None:
        if exception:
            cmd = command.decode("latin-1")
            args = " ".join(str(x) for x in args)
            msg = "Command # {} ({} {}) of pipeline caused error: {}".format(
                number, cmd, args, exception.args[0]
            )
            exception.args = (msg,) + exception.args[1:]

    async def execute(self, raise_on_error: bool = True) -> Tuple[object, ...]:
        await self.connection_pool.initialize()

        if not self.command_stack:
            return ()

        if self._transaction or self.explicit_transaction:
            execute = self.send_cluster_transaction
        else:
            execute = self.send_cluster_commands
        try:
            return await execute(raise_on_error)
        finally:
            await self.reset_pipeline()

    async def reset_pipeline(self):
        """Empties pipeline"""
        self.command_stack = []

        self.scripts: Set[Script[AnyStr]] = set()
        # clean up the other instance attributes
        self.watching = False
        self.explicit_transaction = False
        self._watched_node = None
        if self._watched_connection:
            self.connection_pool.release(self._watched_connection)
            self._watched_connection = None

    @retryable(policy=ConstantRetryPolicy((ClusterDownError,), 3, 0.1))
    async def send_cluster_transaction(
        self, raise_on_error: bool = True
    ) -> Tuple[object, ...]:
        attempt = sorted(self.command_stack, key=lambda x: x.position)
        slots: Set[int] = set()
        for c in attempt:
            slot = self._determine_slot(c.command, *c.args, **c.options)
            if slot:
                slots.add(slot)

            if len(slots) > 1:
                raise ClusterTransactionError("Multiple nodes involved in transaction")
        if not slots:
            raise ClusterTransactionError("No slots found for transaction")
        node = self.connection_pool.get_node_by_slot(slots.pop())

        if self._watched_node and node.name != self._watched_node.name:
            raise ClusterTransactionError("Multiple nodes involved in transaction")

        conn = (
            self._watched_connection
            or await self.connection_pool.get_connection_by_node(node)
        )

        if self.watches:
            await self._watch(node, conn, self.watches)
        node_commands = NodeCommands(
            self.client, conn, in_transaction=True, timeout=self.timeout
        )
        node_commands.append(ClusterPipelineCommand(CommandName.MULTI, ()))
        node_commands.extend(attempt)
        node_commands.append(ClusterPipelineCommand(CommandName.EXEC, ()))
        self.explicit_transaction = True

        await node_commands.write()
        try:
            await node_commands.read()
        except ExecAbortError:
            if self.explicit_transaction:
                request = await conn.create_request(CommandName.DISCARD)
                await request
        # If at least one watched key is modified before the EXEC command,
        # the whole transaction aborts,
        # and EXEC returns a Null reply to notify that the transaction failed.

        if node_commands.commands[-1].result is None:
            raise WatchError
        self.connection_pool.release(conn)

        if self.watching:
            await self._unwatch(conn)

        if raise_on_error:
            self.raise_first_error()

        return tuple(
            n.result
            for n in node_commands.commands
            if n.command not in {CommandName.MULTI, CommandName.EXEC}
        )

    @retryable(policy=ConstantRetryPolicy((ClusterDownError,), 3, 0.1))
    async def send_cluster_commands(
        self, raise_on_error: bool = True, allow_redirections: bool = True
    ) -> Tuple[object, ...]:
        """
        Sends a bunch of cluster commands to the redis cluster.

        `allow_redirections` If the pipeline should follow `ASK` & `MOVED` responses
        automatically. If set to false it will raise RedisClusterException.
        """
        # the first time sending the commands we send all of the commands that were queued up.
        # if we have to run through it again, we only retry the commands that failed.
        attempt = sorted(self.command_stack, key=lambda x: x.position)

        protocol_version: int = 3
        # build a list of node objects based on node names we need to
        nodes: Dict[str, NodeCommands] = {}
        # as we move through each command that still needs to be processed,
        # we figure out the slot number that command maps to, then from the slot determine the node.
        for c in attempt:
            # refer to our internal node -> slot table that tells us where a given
            # command should route to.
            slot = self._determine_slot(c.command, *c.args)
            node = self.connection_pool.get_node_by_slot(slot)

            if node.name not in nodes:
                nodes[node.name] = NodeCommands(
                    self.client,
                    await self.connection_pool.get_connection_by_node(node),
                    timeout=self.timeout,
                )

            nodes[node.name].append(c)

        # send the commands in sequence.
        # we  write to all the open sockets for each node first, before reading anything
        # this allows us to flush all the requests out across the network essentially in parallel
        # so that we can read them all in parallel as they come back.
        # we dont' multiplex on the sockets as they come available, but that shouldn't make
        # too much difference.
        node_commands = nodes.values()

        for n in node_commands:
            await n.write()

        for n in node_commands:
            await n.read()

        # release all of the redis connections we allocated earlier back into the connection pool.
        # we used to do this step as part of a try/finally block, but it is really dangerous to
        # release connections back into the pool if for some reason the socket has data still left
        # in it from a previous operation. The write and read operations already have try/catch
        # around them for all known types of errors including connection and socket level errors.
        # So if we hit an exception, something really bad happened and putting any of
        # these connections back into the pool is a very bad idea.
        # the socket might have unread buffer still sitting in it, and then the
        # next time we read from it we pass the buffered result back from a previous
        # command and every single request after to that connection will always get
        # a mismatched result. (not just theoretical, I saw this happen on production x.x).
        for n in nodes.values():
            protocol_version = n.connection.protocol_version
            self.connection_pool.release(n.connection)
        # if the response isn't an exception it is a valid response from the node
        # we're all done with that command, YAY!
        # if we have more commands to attempt, we've run into problems.
        # collect all the commands we are allowed to retry.
        # (MOVED, ASK, or connection errors or timeout errors)
        attempt = sorted(
            (c for c in attempt if isinstance(c.result, ERRORS_ALLOW_RETRY)),
            key=lambda x: x.position,
        )

        if attempt and allow_redirections:
            # RETRY MAGIC HAPPENS HERE!
            # send these remaing comamnds one at a time using `execute_command`
            # in the main client. This keeps our retry logic in one place mostly,
            # and allows us to be more confident in correctness of behavior.
            # at this point any speed gains from pipelining have been lost
            # anyway, so we might as well make the best attempt to get the correct
            # behavior.
            #
            # The client command will handle retries for each individual command
            # sequentially as we pass each one into `execute_command`. Any exceptions
            # that bubble out should only appear once all retries have been exhausted.
            #
            # If a lot of commands have failed, we'll be setting the
            # flag to rebuild the slots table from scratch. So MOVED errors should
            # correct .commandsthemselves fairly quickly.
            await self.connection_pool.nodes.increment_reinitialize_counter(
                len(attempt)
            )

            for c in attempt:
                try:
                    # send each command individually like we do in the main client.
                    c.result = await self.client.execute_command(
                        c.command, *c.args, **c.options
                    )
                except RedisError as e:
                    c.result = e

        # turn the response back into a simple flat array that corresponds
        # to the sequence of commands issued in the stack in pipeline.execute()
        response = []
        for c in sorted(self.command_stack, key=lambda x: x.position):
            r = c.result
            if not isinstance(c.result, RedisError):
                if isinstance(c.callback, AsyncPreProcessingCallback):
                    await c.callback.pre_process(self.client, c.result, **c.options)
                r = c.callback(c.result, version=protocol_version, **c.options)
            response.append(r)

        if raise_on_error:
            self.raise_first_error()

        return tuple(response)

    def _determine_slot(self, command: bytes, *args: ValueT, **options: ValueT) -> int:
        """Figure out what slot based on command and args"""

        keys: Tuple[ValueT, ...] = cast(
            Tuple[ValueT, ...], options.get("keys")
        ) or KeySpec.extract_keys(command, *args)

        if not keys:
            raise RedisClusterException(
                f"No way to dispatch {command} to Redis Cluster. Missing key"
            )
        slots = {hash_slot(b(key)) for key in keys}

        if len(slots) != 1:
            raise ClusterCrossSlotError(command=command, keys=keys)
        return slots.pop()

    def _fail_on_redirect(self, allow_redirections: bool) -> None:
        if not allow_redirections:
            raise RedisClusterException(
                "ASK & MOVED redirection not allowed in this pipeline"
            )

    def multi(self) -> None:
        if self.explicit_transaction:
            raise RedisError("Cannot issue nested calls to MULTI")

        if self.command_stack:
            raise RedisError(
                "Commands without an initial WATCH have already been issued"
            )
        self.explicit_transaction = True

    async def immediate_execute_command(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., Any] = NoopCallback(),
        **kwargs: Optional[ValueT],
    ) -> Any:
        slot = self._determine_slot(command, *args)
        node = self.connection_pool.get_node_by_slot(slot)
        if command == CommandName.WATCH:
            if self._watched_node and node.name != self._watched_node.name:
                raise ClusterTransactionError(
                    "Cannot issue a watch on a different node in the same transaction"
                )
            else:
                self._watched_node = node
            self._watched_connection = conn = (
                self._watched_connection
                or await self.connection_pool.get_connection_by_node(node)
            )
        else:
            conn = await self.connection_pool.get_connection_by_node(node)

        try:
            request = await conn.create_request(
                command, *args, decode=kwargs.get("decode")
            )

            return callback(
                await request,
                version=conn.protocol_version,
                **kwargs,
            )
        except (ConnectionError, TimeoutError):
            conn.disconnect()

            try:
                if not self.watching:
                    request = await conn.create_request(
                        command, *args, decode=kwargs.get("decode")
                    )
                    return callback(
                        await request, version=conn.protocol_version, **kwargs
                    )
            except ConnectionError:
                # the retry failed so cleanup.
                conn.disconnect()
                await self.reset_pipeline()
                raise
        finally:
            if command in UNWATCH_COMMANDS:
                self.watching = False
            elif command == CommandName.WATCH:
                self.watching = True
                # don't release the connection if the command was a watch
                return
            self.connection_pool.release(conn)

    def load_scripts(self):
        raise RedisClusterException("method load_scripts() is not implemented")

    async def _watch(
        self, node: ManagedNode, conn: BaseConnection, keys: Parameters[KeyT]
    ) -> bool:
        "Watches the values at keys ``keys``"

        for key in keys:
            slot = self._determine_slot(CommandName.WATCH, key)
            dist_node = self.connection_pool.get_node_by_slot(slot)

            if node.name != dist_node.name:
                raise ClusterTransactionError(
                    "Keys in request don't hash to the same node"
                )

        if self.explicit_transaction:
            raise RedisError("Cannot issue a WATCH after a MULTI")
        request = await conn.create_request(CommandName.WATCH, *keys)

        return SimpleStringCallback()(
            cast(StringT, await request),
            version=conn.protocol_version,
        )

    async def _unwatch(self, conn: BaseConnection) -> bool:
        """Unwatches all previously specified keys"""
        request = await conn.create_request(CommandName.UNWATCH, decode=False)
        res = cast(StringT, await request)
        return res == b"OK" if self.watching else True


class Pipeline(ObjectProxy, Generic[AnyStr]):  # type: ignore
    """
    Class returned by :meth:`coredis.Redis.pipeline`

    The class exposes the redis command methods available in
    :class:`~coredis.Redis`, however each of those methods returns
    the instance itself and the results of the batched commands
    can be retrieved by calling :meth:`execute`.
    """

    __wrapped__: PipelineImpl[AnyStr]

    async def __aenter__(self) -> Pipeline[AnyStr]:
        return cast(Pipeline[AnyStr], await self.__wrapped__.__aenter__())

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.__wrapped__.__aexit__(exc_type, exc_value, traceback)

    @classmethod
    def proxy(
        cls,
        client: Redis[AnyStr],
        transaction: Optional[bool] = None,
        watches: Optional[Parameters[KeyT]] = None,
        timeout: Optional[float] = None,
    ) -> Pipeline[AnyStr]:
        return cls(
            PipelineImpl(
                client,
                transaction=transaction,
                watches=watches,
                timeout=timeout,
            )
        )

    def multi(self) -> None:
        """
        Starts a transactional block of the pipeline after WATCH commands
        are issued. End the transactional block with :meth:`execute`
        """
        self.__wrapped__.multi()  # Only here for documentation purposes.

    async def watch(self, *keys: KeyT) -> bool:  # noqa
        """
        Watches the values at keys ``keys``
        """
        return await self.__wrapped__.watch(
            *keys
        )  # Only here for documentation purposes.

    async def unwatch(self) -> bool:  # noqa
        """
        Unwatches all previously specified keys
        """
        return await self.__wrapped__.unwatch()  # Only here for documentation purposes.

    async def execute(self, raise_on_error: bool = True) -> Tuple[object, ...]:
        """
        Executes all the commands in the current pipeline
        and return the results of the individual batched commands
        """

        # Only here for documentation purposes.
        return await self.__wrapped__.execute(raise_on_error=raise_on_error)

    async def reset(self) -> None:
        """
        Resets the command stack and releases any connections acquired from the
        pool
        """
        await self.__wrapped__.reset_pipeline()


class ClusterPipeline(ObjectProxy, Generic[AnyStr]):  # type: ignore
    """
    Class returned by :meth:`coredis.RedisCluster.pipeline`

    The class exposes the redis command methods available in
    :class:`~coredis.Redis`, however each of those methods returns
    the instance itself and the results of the batched commands
    can be retrieved by calling :meth:`execute`.
    """

    __wrapped__: ClusterPipelineImpl[AnyStr]

    async def __aenter__(self) -> ClusterPipeline[AnyStr]:
        return cast(ClusterPipeline[AnyStr], await self.__wrapped__.__aenter__())

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.__wrapped__.__aexit__(exc_type, exc_value, traceback)

    @classmethod
    def proxy(
        cls,
        client: RedisCluster[AnyStr],
        transaction: Optional[bool] = False,
        watches: Optional[Parameters[KeyT]] = None,
        timeout: Optional[float] = None,
    ) -> ClusterPipeline[AnyStr]:
        return cls(
            ClusterPipelineImpl(
                client,
                transaction=transaction,
                watches=watches,
                timeout=timeout,
            )
        )

    def multi(self) -> None:
        """
        Starts a transactional block of the pipeline after WATCH commands
        are issued. End the transactional block with :meth:`execute`
        """
        self.__wrapped__.multi()  # Only here for documentation purposes.

    async def watch(self, *keys: KeyT) -> bool:  # noqa
        """
        Watches the values at keys ``keys``

        :raises: :exc:`~coredis.exceptions.ClusterTransactionError`
         if a watch is issued on a key that resides on a different
         cluster node than a previous watch.
        """
        return await self.__wrapped__.watch(
            *keys
        )  # Only here for documentation purposes.

    async def unwatch(self) -> bool:  # noqa
        """
        Unwatches all previously specified keys
        """
        return await self.__wrapped__.unwatch()  # Only here for documentation purposes.

    async def execute(self, raise_on_error: bool = True) -> Tuple[object, ...]:
        """
        Executes all the commands in the current pipeline
        and return the results of the individual batched commands
        """
        # Only here for documentation purposes.
        return await self.__wrapped__.execute(raise_on_error=raise_on_error)

    async def reset(self) -> None:
        """
        Resets the command stack and releases any connections acquired from the
        pool
        """
        await self.__wrapped__.reset_pipeline()



from __future__ import annotations

import os


class __Config:
    def __init__(self) -> None:
        self.__optimized: bool = False

    @property
    def runtime_checks(self) -> bool:
        """
        Whether runtime type checks are to be enabled.
        Can be enabled by setting the environment variable ``COREDIS_RUNTIME_CHECKS`` to ``true``
        """
        return os.environ.get("COREDIS_RUNTIME_CHECKS", "").lower() in [
            "1",
            "true",
            "t",
        ]

    @property
    def optimized(self) -> bool:
        """
        When ``optimized`` is ``True`` most runtime validations will be disabled.
        This can be enabled in any of the following ways:

          - By running python in optimized mode using the ``-O`` flag
          - By setting the environment variable ``COREDIS_OPTIMIZED`` to ``true``
          - By explicitly setting ``coredis.Config.optimized = True``

        """
        return (
            not __debug__
            or os.environ.get("COREDIS_OPTIMIZED", "").lower()
            in [
                "1",
                "true",
                "t",
            ]
            or self.__optimized
        )

    @optimized.setter
    def optimized(self, value: bool) -> None:
        self.__optimized = value


#: Used to configure global behaviors of the coredis library
Config = __Config()



from __future__ import annotations

import random
import ssl
import weakref
from typing import Any, cast, overload

from coredis import Redis
from coredis._utils import nativestr
from coredis.cache import AbstractCache
from coredis.connection import Connection
from coredis.exceptions import (
    ConnectionError,
    PrimaryNotFoundError,
    ReplicaNotFoundError,
    ResponseError,
    TimeoutError,
)
from coredis.pool import ConnectionPool
from coredis.typing import (
    AnyStr,
    Dict,
    Generic,
    Iterable,
    List,
    Literal,
    Optional,
    StringT,
    Tuple,
    Type,
    Union,
)


class SentinelManagedConnection(Connection, Generic[AnyStr]):
    def __init__(
        self,
        connection_pool: SentinelConnectionPool,
        host: str = "127.0.0.1",
        port: int = 6379,
        username: Optional[str] = None,
        password: Optional[str] = None,
        db: int = 0,
        stream_timeout: Optional[float] = None,
        connect_timeout: Optional[float] = None,
        ssl_context: Optional[ssl.SSLContext] = None,
        encoding: str = "utf-8",
        decode_responses: bool = False,
        socket_keepalive: Optional[bool] = None,
        socket_keepalive_options: Optional[Dict[int, Union[int, bytes]]] = None,
        *,
        client_name: Optional[str] = None,
        protocol_version: Literal[2, 3] = 2,
    ):
        self.connection_pool: SentinelConnectionPool = weakref.proxy(connection_pool)
        super().__init__(
            host=host,
            port=port,
            username=username,
            password=password,
            db=db,
            stream_timeout=stream_timeout,
            connect_timeout=connect_timeout,
            ssl_context=ssl_context,
            encoding=encoding,
            decode_responses=decode_responses,
            socket_keepalive=socket_keepalive,
            socket_keepalive_options=socket_keepalive_options,
            client_name=client_name,
            protocol_version=protocol_version,
        )

    def __repr__(self) -> str:
        pool = self.connection_pool

        if self.host:
            host_info = f",host={self.host},port={self.port}"
        else:
            host_info = ""
        s = f"{type(self).__name__}<service={pool.service_name}{host_info}>"

        return s

    async def connect_to(self, address: Tuple[str, int]) -> None:
        self.host, self.port = address
        await super().connect()

    async def connect(self) -> None:
        if not self.is_connected:
            if self.connection_pool.is_primary:
                await self.connect_to(await self.connection_pool.get_primary_address())
            else:
                for replica in await self.connection_pool.rotate_replicas():
                    try:
                        return await self.connect_to(replica)
                    except ConnectionError:
                        continue
                raise ReplicaNotFoundError  # Never be here

        return None


class SentinelConnectionPool(ConnectionPool):
    """
    Sentinel backed connection pool.
    """

    primary_address: Optional[Tuple[str, int]]
    replica_counter: Optional[int]

    def __init__(
        self,
        service_name: StringT,
        sentinel_manager: Sentinel[Any],
        is_primary: bool = True,
        check_connection: bool = True,
        **kwargs: Any,
    ):
        self.is_primary = is_primary
        kwargs["connection_class"] = cast(
            Type[Connection],
            kwargs.get(
                "connection_class", SentinelManagedConnection[AnyStr]  # type: ignore
            ),
        )
        super().__init__(**kwargs)
        self.connection_kwargs["connection_pool"] = self
        self.service_name = nativestr(service_name)
        self.sentinel_manager = sentinel_manager
        self.check_connection = check_connection

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}"
            f"<service={self.service_name}"
            f"({'primary' if self.is_primary else 'replica'})"
        )

    def reset(self) -> None:
        super().reset()
        self.primary_address = None
        self.replica_counter = None

    async def get_primary_address(self) -> Tuple[str, int]:
        primary_address = await self.sentinel_manager.discover_primary(
            self.service_name
        )

        if self.is_primary:
            if self.primary_address is None:
                self.primary_address = primary_address
            elif primary_address != self.primary_address:
                # Primary address changed, disconnect all clients in this pool
                self.disconnect()

        return primary_address

    async def rotate_replicas(self) -> List[Tuple[str, int]]:
        """Round-robin replicas balancer"""
        replicas = await self.sentinel_manager.discover_replicas(self.service_name)
        replica_addresses: List[Tuple[str, int]] = []

        if replicas:
            if self.replica_counter is None:
                self.replica_counter = random.randint(0, len(replicas) - 1)

            for _ in range(len(replicas)):
                self.replica_counter = (self.replica_counter + 1) % len(replicas)
                replica_addresses.append(replicas[self.replica_counter])

            return replica_addresses
        # Fallback to primary
        try:
            return [await self.get_primary_address()]
        except PrimaryNotFoundError:
            pass
        raise ReplicaNotFoundError("No replica found for %r" % (self.service_name))


class Sentinel(Generic[AnyStr]):
    """
    Example use::

        from coredis.sentinel import Sentinel
        sentinel = Sentinel([('localhost', 26379)], stream_timeout=0.1)
        async def test():
            primary = await sentinel.primary_for('my-instance', stream_timeout=0.1)
            await primary.set('foo', 'bar')
            replica = await sentinel.replica_for('my-instance', stream_timeout=0.1)
            await replica.get('foo')

    """

    @overload
    def __init__(
        self: Sentinel[bytes],
        sentinels: Iterable[Tuple[str, int]],
        min_other_sentinels: int = ...,
        sentinel_kwargs: Optional[Dict[str, Any]] = ...,
        decode_responses: Literal[False] = ...,
        cache: Optional[AbstractCache] = None,
        **connection_kwargs: Any,
    ) -> None: ...

    @overload
    def __init__(
        self: Sentinel[str],
        sentinels: Iterable[Tuple[str, int]],
        min_other_sentinels: int = ...,
        sentinel_kwargs: Optional[Dict[str, Any]] = ...,
        decode_responses: Literal[True] = ...,
        cache: Optional[AbstractCache] = None,
        **connection_kwargs: Any,
    ) -> None: ...

    def __init__(
        self,
        sentinels: Iterable[Tuple[str, int]],
        min_other_sentinels: int = 0,
        sentinel_kwargs: Optional[Dict[str, Any]] = None,
        decode_responses: bool = False,
        cache: Optional[AbstractCache] = None,
        **connection_kwargs: Any,
    ) -> None:
        """
        Changes
          - .. versionadded:: 3.10.0
               Accept :paramref:`cache` parameter to be used with primaries
               and replicas returned from the sentinel instance.

        :param sentinels: is a list of sentinel nodes. Each node is represented by
         a pair (hostname, port).
        :param min_other_sentinels: defined a minimum number of peers for a sentinel.
         When querying a sentinel, if it doesn't meet this threshold, responses
         from that sentinel won't be considered valid.
        :param sentinel_kwargs: is a dictionary of connection arguments used when
         connecting to sentinel instances. Any argument that can be passed to
         a normal Redis connection can be specified here. If :paramref:`sentinel_kwargs` is
         not specified, ``stream_timeout``, ``socket_keepalive``, ``decode_responses``
         and ``protocol_version`` options specified in :paramref:`connection_kwargs` will be used.
        :param cache: If provided the cache will be shared between both primaries and replicas
         returned by this sentinel.
        :param connection_kwargs: are keyword arguments that will be used when
         establishing a connection to a Redis server (i.e. are passed on to the
         constructor of :class:`Redis` for all primary and replicas).
        """
        # if sentinel_kwargs isn't defined, use the socket_* options from
        # connection_kwargs

        if not sentinel_kwargs:
            sentinel_kwargs = {
                k: v
                for k, v in iter(connection_kwargs.items())
                if k
                in {
                    "socket_timeout",
                    "socket_keepalive",
                    "encoding",
                    "protocol_version",
                }
            }

        self.sentinel_kwargs = sentinel_kwargs
        self.min_other_sentinels = min_other_sentinels
        self.connection_kwargs = connection_kwargs
        self.__cache = cache
        self.connection_kwargs["decode_responses"] = self.sentinel_kwargs[
            "decode_responses"
        ] = decode_responses

        self.sentinels = [
            Redis(hostname, port, **self.sentinel_kwargs)
            for hostname, port in sentinels
        ]

    def __repr__(self) -> str:
        sentinel_addresses: List[str] = []

        for sentinel in self.sentinels:
            sentinel_addresses.append(
                "{}:{}".format(
                    sentinel.connection_pool.connection_kwargs["host"],
                    sentinel.connection_pool.connection_kwargs["port"],
                )
            )

        return "{}<sentinels=[{}]>".format(
            type(self).__name__, ",".join(sentinel_addresses)
        )

    def __check_primary_state(
        self,
        state: Dict[str, Union[int, bool, str]],
    ) -> bool:
        if not state["is_master"] or state["is_sdown"] or state["is_odown"]:
            return False

        if int(state["num-other-sentinels"]) < self.min_other_sentinels:
            return False

        return True

    def __filter_replicas(
        self, replicas: Iterable[Dict[str, Union[str, int, bool]]]
    ) -> List[Tuple[str, int]]:
        """Removes replicas that are in an ODOWN or SDOWN state"""
        replicas_alive: List[Tuple[str, int]] = []

        for replica in replicas:
            if replica["is_odown"] or replica["is_sdown"]:
                continue
            replicas_alive.append((nativestr(replica["ip"]), int(replica["port"])))

        return replicas_alive

    async def discover_primary(self, service_name: str) -> Tuple[str, int]:
        """
        Asks sentinel servers for the Redis primary's address corresponding
        to the service labeled :paramref:`service_name`.

        :return: A pair (address, port) or raises :exc:`~coredis.exceptions.PrimaryNotFoundError`
         if no primary is found.
        """

        for sentinel_no, sentinel in enumerate(self.sentinels):
            try:
                primaries = await sentinel.sentinel_masters()
            except (ConnectionError, TimeoutError):
                continue
            state = primaries.get(service_name)

            if state and self.__check_primary_state(state):
                # Put this sentinel at the top of the list
                self.sentinels[0], self.sentinels[sentinel_no] = (
                    sentinel,
                    self.sentinels[0],
                )

                return nativestr(state["ip"]), int(state["port"])
        raise PrimaryNotFoundError(f"No primary found for {service_name!r}")

    async def discover_replicas(self, service_name: str) -> List[Tuple[str, int]]:
        """Returns a list of alive replicas for service :paramref:`service_name`"""

        for sentinel in self.sentinels:
            try:
                replicas = await sentinel.sentinel_replicas(service_name)
            except (ConnectionError, ResponseError, TimeoutError):
                continue
            filtered_replicas = self.__filter_replicas(replicas)

            if filtered_replicas:
                return filtered_replicas

        return []

    @overload
    def primary_for(
        self: Sentinel[bytes],
        service_name: str,
        *,
        redis_class: Type[Redis[bytes]] = ...,
        connection_pool_class: Type[SentinelConnectionPool] = ...,
        **kwargs: Any,
    ) -> Redis[bytes]: ...

    @overload
    def primary_for(
        self: Sentinel[str],
        service_name: str,
        *,
        redis_class: Type[Redis[str]] = ...,
        connection_pool_class: Type[SentinelConnectionPool] = ...,
        **kwargs: Any,
    ) -> Redis[str]: ...

    def primary_for(
        self,
        service_name: str,
        *,
        redis_class: Type[Redis[Any]] = Redis[Any],
        connection_pool_class: Type[SentinelConnectionPool] = SentinelConnectionPool,
        **kwargs: Any,
    ) -> Union[Redis[bytes], Redis[str]]:
        """
        Returns a redis client instance for the :paramref:`service_name` primary.

        A :class:`coredis.sentinel.SentinelConnectionPool` class is used to
        retrive the primary's address before establishing a new connection.

        NOTE: If the primary's address has changed, any cached connections to
        the old primary are closed.

        By default clients will be a :class:`~coredis.Redis` instances.
        Specify a different class to the :paramref:`redis_class` argument if you desire
        something different.

        The :paramref:`connection_pool_class` specifies the connection pool to use.
        The :class:`~coredis.sentinel.SentinelConnectionPool` will be used by default.

        All other keyword arguments are merged with any :paramref:`Sentinel.connection_kwargs`
        passed to this class and passed to the connection pool as keyword
        arguments to be used to initialize Redis connections.
        """
        kwargs["is_primary"] = True
        connection_kwargs = dict(self.connection_kwargs)
        connection_kwargs.update(kwargs)

        return redis_class(
            connection_pool=connection_pool_class(
                service_name,
                self,
                **connection_kwargs,
            ),
            cache=self.__cache,
        )

    @overload
    def replica_for(
        self: Sentinel[bytes],
        service_name: str,
        redis_class: Type[Redis[bytes]] = ...,
        connection_pool_class: Type[SentinelConnectionPool] = ...,
        **kwargs: Any,
    ) -> Redis[bytes]: ...

    @overload
    def replica_for(
        self: Sentinel[str],
        service_name: str,
        redis_class: Type[Redis[str]] = ...,
        connection_pool_class: Type[SentinelConnectionPool] = ...,
        **kwargs: Any,
    ) -> Redis[str]: ...

    def replica_for(
        self,
        service_name: str,
        redis_class: Type[Redis[Any]] = Redis[Any],
        connection_pool_class: Type[SentinelConnectionPool] = SentinelConnectionPool,
        **kwargs: Any,
    ) -> Union[Redis[bytes], Redis[str]]:
        """
        Returns redis client instance for the :paramref:`service_name` replica(s).

        A SentinelConnectionPool class is used to retrieve the replica's
        address before establishing a new connection.

        By default clients will be a redis.Redis instance. Specify a
        different class to the :paramref:`redis_class` argument if you desire
        something different.

        The :paramref:`connection_pool_class` specifies the connection pool to use.
        The SentinelConnectionPool will be used by default.

        All other keyword arguments are merged with any :paramref:`Sentinel.connection_kwargs`
        passed to this class and passed to the connection pool as keyword
        arguments to be used to initialize Redis connections.
        """
        kwargs["is_primary"] = False
        connection_kwargs = dict(self.connection_kwargs)
        connection_kwargs.update(kwargs)

        return redis_class(
            connection_pool=connection_pool_class(
                service_name,
                self,
                **connection_kwargs,
            ),
            cache=self.__cache,
        )



"""
coredis.commands
----------------
Implementation of core redis commands and abstractions over high level
core concepts such as pubsub, scripting and functions.
"""

from __future__ import annotations

from abc import ABC, abstractmethod

from coredis.response._callbacks import NoopCallback
from coredis.typing import AnyStr, Callable, Generic, Optional, R, ValueT

# Command wrappers
from .bitfield import BitFieldOperation
from .function import Function, Library
from .monitor import Monitor
from .pubsub import ClusterPubSub, PubSub, ShardedPubSub
from .script import Script


class CommandMixin(Generic[AnyStr], ABC):
    @abstractmethod
    async def execute_command(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., R] = NoopCallback(),
        **options: Optional[ValueT],
    ) -> R:
        pass


__all__ = [
    "BitFieldOperation",
    "ClusterPubSub",
    "Function",
    "Library",
    "Monitor",
    "PubSub",
    "Script",
    "ShardedPubSub",
]



from __future__ import annotations

import datetime
import time
import warnings
from typing import TYPE_CHECKING, Any

from coredis.commands.constants import CommandName
from coredis.config import Config
from coredis.exceptions import CommandNotSupportedError, CommandSyntaxError
from coredis.typing import Dict, Optional, Union

if TYPE_CHECKING:
    import coredis.client
    from coredis.commands._wrappers import CommandDetails


def normalized_seconds(value: Union[int, datetime.timedelta]) -> int:
    if isinstance(value, datetime.timedelta):
        value = value.seconds + value.days * 24 * 3600

    return value


def normalized_milliseconds(value: Union[int, datetime.timedelta]) -> int:
    if isinstance(value, datetime.timedelta):
        ms = int(value.microseconds / 1000)
        value = (value.seconds + value.days * 24 * 3600) * 1000 + ms

    return value


def normalized_time_seconds(value: Union[int, datetime.datetime]) -> int:
    if isinstance(value, datetime.datetime):
        s = int(value.microsecond / 1000000)
        value = int(time.mktime(value.timetuple())) + s

    return value


def normalized_time_milliseconds(value: Union[int, datetime.datetime]) -> int:
    if isinstance(value, datetime.datetime):
        ms = int(value.microsecond / 1000)
        value = int(time.mktime(value.timetuple())) * 1000 + ms

    return value


async def check_version(
    instance: coredis.client.Client[Any],
    function_name: str,
    command_details: "CommandDetails",
    deprecation_reason: Optional[str] = None,
    kwargs: Dict[str, Any] = {},
) -> None:
    if Config.optimized or not any(
        [
            command_details.version_introduced,
            command_details.version_deprecated,
            command_details.arguments,
        ]
    ):
        return
    if getattr(instance, "verify_version", False) and not getattr(
        instance, "noreply", False
    ):
        server_version = getattr(instance, "server_version", None)
        if not server_version:
            return
        if (
            command_details.version_introduced
            and server_version < command_details.version_introduced
        ):
            raise CommandNotSupportedError(
                command_details.command.decode("latin-1"),
                str(instance.server_version),
            )
        elif command_details.arguments and set(
            command_details.arguments.keys()
        ).intersection(kwargs.keys()):
            for argument, minimum_version in command_details.arguments.items():
                if minimum_version and server_version < minimum_version:
                    raise CommandSyntaxError(
                        {argument},
                        (
                            f"{command_details.command.decode('latin-1')} with `{argument}` "
                            f"is not supported in redis version {server_version}"
                        ),
                    )
        elif (
            command_details.version_deprecated
            and server_version >= command_details.version_deprecated
        ):
            warnings.warn(
                (
                    deprecation_reason.strip()
                    if deprecation_reason
                    else (
                        f"{function_name}() is deprecated since redis version "
                        "{command_details.version_deprecated}."
                    )
                ),
                category=DeprecationWarning,
                stacklevel=3,
            )


def redis_command_link(command: CommandName) -> str:
    return f'`{str(command)} <https://redis.io/commands/{str(command).lower().replace(" ", "-")}>`_'



from __future__ import annotations

import dataclasses
import itertools
from datetime import timedelta

from deprecated.sphinx import versionadded

from ..commands._utils import normalized_milliseconds, normalized_seconds
from ..commands._wrappers import ClusterCommandConfig
from ..commands.constants import CommandGroup, CommandName, NodeFlag
from ..response._callbacks import (
    AnyStrCallback,
    ClusterEnsureConsistent,
    ClusterMergeSets,
    DictCallback,
    IntCallback,
    SetCallback,
    SimpleStringCallback,
)
from ..tokens import PrefixToken, PureToken
from ..typing import (
    AnyStr,
    CommandArgList,
    Dict,
    KeyT,
    List,
    Literal,
    Mapping,
    Optional,
    Parameters,
    ResponsePrimitive,
    ResponseType,
    Set,
    StringT,
    Tuple,
    Union,
    ValueT,
)
from .base import Module, ModuleGroup, module_command
from .response._callbacks.search import (
    AggregationResultCallback,
    SearchConfigCallback,
    SearchResultCallback,
    SpellCheckCallback,
    SpellCheckResult,
)
from .response.types import SearchAggregationResult, SearchResult


class RediSearch(Module[AnyStr]):
    NAME = "search"
    FULL_NAME = "RediSearch"
    DESCRIPTION = """RedisSearch is a Redis module that enables querying, secondary 
indexing, and full-text search for Redis. These features enable multi-field queries, 
aggregation, exact phrase matching, numeric filtering, geo filtering and vector 
similarity semantic search on top of text queries."""
    DOCUMENTATION_URL = "https://redis.io/docs/stack/search"


@dataclasses.dataclass
class Field:
    """
    Field definition to be used in :meth:`~coredis.modules.Search.create` &
    :meth:`~coredis.modules.Search.alter`

    For more details refer to the documentation for
    `FT.CREATE <https://redis.io/commands/ft.create/>`__
    """

    #: Name of the field. For hashes, this is a field name within the hash.
    #:  For JSON, this is a JSON Path expression.
    name: StringT
    #: Type of field
    type: Literal[
        PureToken.TEXT,
        PureToken.TAG,
        PureToken.NUMERIC,
        PureToken.GEO,
        PureToken.VECTOR,
    ]

    #: Defines the alias associated to :paramref:`name`.
    #: For example, you can use this feature to alias a complex
    #: JSONPath expression with more memorable (and easier to type) name.
    alias: Optional[StringT] = None
    #: Whether to optimize for sorting.
    sortable: Optional[bool] = None
    #: Whether to use the unnormalized form of the field for sorting.
    unf: Optional[bool] = None
    #: Whether to disable stemming for this field.
    nostem: Optional[bool] = None
    #: Skip indexing this field
    noindex: Optional[bool] = None
    #: Phonetic algorithm to use for this field.
    phonetic: Optional[StringT] = None
    #: Weight of this field in the document's ranking. The default is 1.0.
    weight: Optional[Union[int, float]] = None
    #: Separator to use for splitting tags if the field is of
    #: type :attr:`~coredis.PureToken.TAG`.
    separator: Optional[StringT] = None
    #: For fields of type :attr:`~coredis.PureToken.TAG`,
    #: keeps the original letter cases of the tags. If not specified,
    #: the characters are converted to lowercase.
    casesensitive: Optional[bool] = None
    #: For fields of type :attr:`~coredis.PureToken.TAG` &
    #: :attr:`~coredis.PureToken.TEXT`, keeps a suffix trie with all
    #: terms which match the suffix. It is used to optimize contains ``(foo)``
    #: and suffix ``(*foo)`` queries. Otherwise, a brute-force search on the trie
    #: is performed. If suffix trie exists for some fields, these queries will
    #: be disabled for other fields
    withsuffixtrie: Optional[bool] = None
    #: The algorithm to use for indexing if the field is of type
    #: :attr:`~coredis.PureToken.VECTOR`.
    #: For more details refer to the
    #: `Vector similarity <https://redis.io/docs/stack/search/reference/vectors/>`__
    #: section of the RediSearch documentation.
    algorithm: Optional[Literal["FLAT", "HSNW"]] = None
    #: A dictionary of attributes to be used with the :paramref:`algorithm` specified.
    #: For more details refer to the
    #: `Creation attributes per algorithm <https://redis.io/docs/stack/search/reference/vectors/#creation-attributes-per-algorithm>`__
    #: section of the RediSearch documentation.
    attributes: Optional[Dict[StringT, ValueT]] = None

    @property
    def args(self) -> Tuple[ValueT, ...]:
        args: CommandArgList = [self.name]
        if self.alias:
            args += [PrefixToken.AS, self.alias]
        args += [self.type]
        if self.type == PureToken.VECTOR:
            assert self.algorithm

            args += [self.algorithm]
            if self.attributes:
                _attributes: List[ValueT] = list(
                    itertools.chain(*self.attributes.items())
                )
                args += [len(_attributes)]
                args += _attributes

        if self.sortable:
            args += [PureToken.SORTABLE]
        if self.unf:
            args += [PureToken.UNF]
        if self.nostem:
            args += [b"NOSTEM"]
        if self.noindex:
            args += [PureToken.NOINDEX]
        if self.phonetic:
            args += [b"PHONETIC", self.phonetic]
        if self.weight:
            args += [b"WEIGHT", self.weight]
        if self.separator:
            args += [PrefixToken.SEPARATOR, self.separator]
        if self.casesensitive:
            args += [b"CASESENSITIVE"]
        if self.withsuffixtrie:
            args += [PureToken.WITHSUFFIXTRIE]
        return tuple(args)


@dataclasses.dataclass
class Reduce:
    """
    Reduce definition to be used with :paramref:`~coredis.modules.Search.aggregate.transformations`
    to define ``REDUCE`` steps in :meth:`~coredis.modules.Search.aggregate`

    For more details refer to `GroupBy Reducers <https://redis.io/docs/stack/search/reference/aggregations/#groupby-reducers>`__
    in the RediSearch documentation.
    """

    #: The name of the reducer function
    function: StringT
    #: The arguments to the reducer function
    parameters: Optional[Parameters[ValueT]] = None
    #: The alias to assign to the result of the reducer function
    alias: Optional[StringT] = None

    @property
    def args(self) -> CommandArgList:
        args: CommandArgList = [self.function]
        if self.parameters:
            args.extend(self.parameters)
        if self.alias:
            args.extend([PrefixToken.AS, self.alias])
        return args


@dataclasses.dataclass
class Group:
    """
    Group definition to be used with :paramref:`~coredis.modules.Search.aggregate.transformations`
    to specify ``GROUPBY`` steps in :meth:`~coredis.modules.Search.aggregate`

    For more details refer to
    `Aggregations <https://redis.io/docs/stack/search/reference/aggregations>`__
    in the RediSearch documentation.
    """

    #: The field to group by
    by: Union[StringT, Parameters[StringT]]
    #: The reducers to apply to each group
    reducers: Optional[Parameters[Reduce]] = None

    @property
    def args(self) -> CommandArgList:
        args: CommandArgList = [PrefixToken.GROUPBY]
        if isinstance(self.by, (bytes, str)):
            args.extend([1, self.by])
        else:
            bies: List[StringT] = list(self.by)
            args.extend([len(bies), *bies])
        for reducer in self.reducers or []:
            args.append(PrefixToken.REDUCE)
            args.extend(reducer.args)

        return args


@dataclasses.dataclass
class Apply:
    """
    Apply definition to be used with :paramref:`~coredis.modules.Search.aggregate.transformations`
    to specify ``APPLY`` steps in :meth:`~coredis.modules.Search.aggregate`

    For more details refer to
    `APPLY expressions <https://redis.io/docs/stack/search/reference/aggregations/#apply-expressions>`__
    in the RediSearch documentation.
    """

    #: The expression to apply
    function: StringT
    #: The alias to assign to the result of the expression
    alias: StringT

    @property
    def args(self) -> CommandArgList:
        return [PrefixToken.APPLY, self.function, PrefixToken.AS, self.alias]


@dataclasses.dataclass
class Filter:
    """
    Filter definition to be used with :paramref:`~coredis.modules.Search.aggregate.transformations`
    to specify ``FILTER`` steps in :meth:`~coredis.modules.Search.aggregate`

    For more details refer to
    `FILTER expressions <https://redis.io/docs/stack/search/reference/aggregations/#filter-expressions>`__
    in the RediSearch documentation.
    """

    #: The filter expression
    expression: StringT

    @property
    def args(self) -> CommandArgList:
        return [PrefixToken.FILTER, self.expression]


@versionadded(version="4.12")
class Search(ModuleGroup[AnyStr]):
    MODULE = RediSearch
    COMMAND_GROUP = CommandGroup.SEARCH

    @module_command(
        CommandName.FT_CREATE,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
    )
    async def create(
        self,
        index: KeyT,
        schema: Parameters[Field],
        *,
        on: Optional[Literal[PureToken.HASH, PureToken.JSON]] = None,
        prefixes: Optional[Parameters[StringT]] = None,
        filter_expression: Optional[StringT] = None,
        language: Optional[StringT] = None,
        language_field: Optional[StringT] = None,
        score: Optional[Union[int, float]] = None,
        score_field: Optional[StringT] = None,
        payload_field: Optional[StringT] = None,
        maxtextfields: Optional[bool] = None,
        temporary: Optional[Union[int, timedelta]] = None,
        nooffsets: Optional[bool] = None,
        nohl: Optional[bool] = None,
        nofields: Optional[bool] = None,
        nofreqs: Optional[bool] = None,
        stopwords: Optional[Parameters[StringT]] = None,
        skipinitialscan: Optional[bool] = None,
    ) -> bool:
        """
        Creates an index with the given spec


        :param index: The name of the index to create.
        :param schema: A list of schema fields
        :param on: The type of Redis key to index.
        :param prefixes: A list of key prefixes to index.
        :param filter_expression: A filter expression to apply to the index.
        :param language: The default language to use for text fields.
        :param language_field: The name of the field to use for language detection.
        :param score: The default score to use for documents.
        :param score_field: The name of the field to use for document scoring.
        :param payload_field: The name of the field to use for document payloads.
        :param maxtextfields: If ``True``, the maximum number of text fields will be used.
        :param temporary: If specified, the index will be temporary and
         will expire after the given number of seconds.
        :param nooffsets: If ``True``, term offsets will not be stored.
        :param nohl: If ``True``, search results will not include highlighted snippets.
        :param nofields: If ``True``, search results will not include field values.
        :param nofreqs: If ``True``, term frequencies will not be stored.
        :param stopwords: A list of stopwords to ignore.
        :param skipinitialscan: If ``True``, the initial scan of the index will be skipped.
        """
        pieces: CommandArgList = [index]
        if on:
            pieces.extend([PrefixToken.ON, on])

        if prefixes:
            _prefixes: List[StringT] = list(prefixes)
            pieces.extend([PrefixToken.PREFIX, len(_prefixes), *_prefixes])
        if filter_expression:
            pieces.extend([PrefixToken.FILTER, filter_expression])
        if language:
            pieces.extend([PrefixToken.LANGUAGE, language])
        if language_field:
            pieces.extend([PrefixToken.LANGUAGE_FIELD, language_field])
        if score:
            pieces.extend([PrefixToken.SCORE, score])
        if score_field:
            pieces.extend([PrefixToken.SCORE_FIELD, score_field])
        if payload_field:
            pieces.extend([PrefixToken.PAYLOAD_FIELD, payload_field])
        if maxtextfields:
            pieces.append(PureToken.MAXTEXTFIELDS)
        if temporary:
            pieces.extend([PrefixToken.TEMPORARY, normalized_seconds(temporary)])
        if nooffsets:
            pieces.append(PureToken.NOOFFSETS)
        if nohl:
            pieces.append(PureToken.NOHL)
        if nofields:
            pieces.append(PureToken.NOFIELDS)
        if nofreqs:
            pieces.append(PureToken.NOFREQS)
        if stopwords:
            _stop: List[StringT] = list(stopwords)
            pieces.extend([PrefixToken.STOPWORDS, len(_stop), *_stop])
        if skipinitialscan:
            pieces.append(PureToken.SKIPINITIALSCAN)

        field_args: CommandArgList = [PureToken.SCHEMA]
        for field in schema:
            field_args.extend(field.args)
        pieces.extend(field_args)

        return await self.execute_module_command(
            CommandName.FT_CREATE, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.FT_INFO,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
    )
    async def info(self, index: KeyT) -> Dict[AnyStr, ResponseType]:
        """
        Returns information and statistics on the index

        :param index: The name of the index.
        """
        return await self.execute_module_command(
            CommandName.FT_INFO,
            index,
            callback=DictCallback[AnyStr, ResponseType](
                recursive=[
                    "attributes",
                    "index_definition",
                    "gc_stats",
                    "cursor_stats",
                    "dialect_stats",
                ]
            ),
        )

    @module_command(
        CommandName.FT_EXPLAIN,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
        arguments={"dialect": {"version_introduced": "2.4.3"}},
    )
    async def explain(
        self, index: KeyT, query: StringT, dialect: Optional[int] = None
    ) -> AnyStr:
        """
        Returns the execution plan for a complex query

        :param index: The name of the index to query.
        :param query: The query to explain.
        :param dialect: Query dialect to use.
        """
        pieces: CommandArgList = [index, query]
        if dialect:
            pieces.extend([PrefixToken.DIALECT, dialect])
        return await self.execute_module_command(
            CommandName.FT_EXPLAIN, *pieces, callback=AnyStrCallback[AnyStr]()
        )

    @module_command(
        CommandName.FT_ALTER,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
    )
    async def alter(
        self,
        index: KeyT,
        field: Field,
        skipinitialscan: Optional[bool] = None,
    ) -> bool:
        """
        Adds a new field to the index

        :param index: The name of the index to alter.
        :param field: The new field to add
        :param skipinitialscan: If ``True``, skip the initial scan and indexing.

        """
        pieces: CommandArgList = [index]
        if skipinitialscan:
            pieces.append(PureToken.SKIPINITIALSCAN)
        pieces.extend([PureToken.SCHEMA, PureToken.ADD, *field.args])

        return await self.execute_module_command(
            CommandName.FT_ALTER, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.FT_DROPINDEX,
        module=MODULE,
        version_introduced="2.0.0",
        group=COMMAND_GROUP,
    )
    async def dropindex(self, index: KeyT, delete_docs: Optional[bool] = None) -> bool:
        """
        Deletes the index

        :param index: The name of the index to delete.
        :param delete_docs: If ``True``, delete the documents associated with the index.
        """
        pieces: CommandArgList = [index]
        if delete_docs:
            pieces.append(PureToken.DELETE_DOCS)

        return await self.execute_module_command(
            CommandName.FT_DROPINDEX, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.FT_ALIASADD,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
    )
    async def aliasadd(self, alias: StringT, index: KeyT) -> bool:
        """
        Adds an alias to the index

        :param alias: The alias to be added to the index.
        :param index: The index to which the alias will be added.
        """

        return await self.execute_module_command(
            CommandName.FT_ALIASADD, alias, index, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.FT_ALIASUPDATE,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
    )
    async def aliasupdate(self, alias: StringT, index: KeyT) -> bool:
        """
        Adds or updates an alias to the index

        :param alias: The alias to be added to an index.
        :param index: The index to which the alias will be added.
        """

        return await self.execute_module_command(
            CommandName.FT_ALIASUPDATE, alias, index, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.FT_ALIASDEL,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
    )
    async def aliasdel(self, alias: StringT) -> bool:
        """
        Deletes an alias from the index

        :param alias: The index alias to be removed.
        """

        return await self.execute_module_command(
            CommandName.FT_ALIASDEL, alias, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.FT_TAGVALS,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
    )
    async def tagvals(self, index: KeyT, field_name: StringT) -> Set[AnyStr]:
        """
        Returns the distinct tags indexed in a Tag field

        :param index: The name of the index.
        :param field_name: Name of a Tag field defined in the schema.
        """

        return await self.execute_module_command(
            CommandName.FT_TAGVALS, index, field_name, callback=SetCallback[AnyStr]()
        )

    @module_command(
        CommandName.FT_SYNUPDATE,
        module=MODULE,
        version_introduced="1.2.0",
        group=COMMAND_GROUP,
    )
    async def synupdate(
        self,
        index: KeyT,
        synonym_group: StringT,
        terms: Parameters[StringT],
        skipinitialscan: Optional[bool] = None,
    ) -> bool:
        """
        Creates or updates a synonym group with additional terms

        :param index: The name of the index.
        :param synonym_group: The ID of the synonym group to update.
        :param terms: A list of terms to add to the synonym group.
        :param skipinitialscan: If ``True``, only documents indexed after the
         update will be affected.

        """
        pieces: CommandArgList = [index, synonym_group]
        if skipinitialscan:
            pieces.append(PureToken.SKIPINITIALSCAN)
        pieces.extend(terms)
        return await self.execute_module_command(
            CommandName.FT_SYNUPDATE, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.FT_SYNDUMP,
        module=MODULE,
        version_introduced="1.2.0",
        group=COMMAND_GROUP,
    )
    async def syndump(self, index: KeyT) -> Dict[AnyStr, List[AnyStr]]:
        """
        Dumps the contents of a synonym group

        :param index: The name of the index.
        """

        return await self.execute_module_command(
            CommandName.FT_SYNDUMP, index, callback=DictCallback[AnyStr, List[AnyStr]]()
        )

    @module_command(
        CommandName.FT_SPELLCHECK,
        module=MODULE,
        version_introduced="1.4.0",
        group=COMMAND_GROUP,
        arguments={"dialect": {"version_introduced": "2.4.3"}},
    )
    async def spellcheck(
        self,
        index: KeyT,
        query: StringT,
        distance: Optional[int] = None,
        include: Optional[StringT] = None,
        exclude: Optional[StringT] = None,
        dialect: Optional[int] = None,
    ) -> SpellCheckResult:
        """
        Performs spelling correction on a query, returning suggestions for misspelled terms

        :param index: The name of the index with the indexed terms.
        :param query: The search query.
        :param distance: Maximum Levenshtein distance for spelling suggestions
        :param include: Specifies an inclusion of a custom dictionary
        :param exclude: Specifies an exclusion of a custom dictionary
        :param dialect: The query dialect to use.
        """
        pieces: CommandArgList = [index, query]
        if distance:
            pieces.extend([PrefixToken.DISTANCE, distance])
        if exclude:
            pieces.extend([PrefixToken.TERMS, PureToken.EXCLUDE, exclude])
        if include:
            pieces.extend([PrefixToken.TERMS, PureToken.INCLUDE, include])
        if dialect:
            pieces.extend([PrefixToken.DIALECT, dialect])
        return await self.execute_module_command(
            CommandName.FT_SPELLCHECK, *pieces, callback=SpellCheckCallback()
        )

    @module_command(
        CommandName.FT_DICTADD,
        module=MODULE,
        version_introduced="1.4.0",
        group=COMMAND_GROUP,
    )
    async def dictadd(
        self,
        name: StringT,
        terms: Parameters[StringT],
    ) -> int:
        """
        Adds terms to a dictionary

        :param name: The name of the dictionary.
        :param terms: The terms to add to the dictionary.
        """
        pieces: CommandArgList = [name, *terms]

        return await self.execute_module_command(
            CommandName.FT_DICTADD, *pieces, callback=IntCallback()
        )

    @module_command(
        CommandName.FT_DICTDEL,
        module=MODULE,
        version_introduced="1.4.0",
        group=COMMAND_GROUP,
    )
    async def dictdel(
        self,
        name: StringT,
        terms: Parameters[StringT],
    ) -> int:
        """
        Deletes terms from a dictionary

        :param name: The name of the dictionary.
        :param terms: The terms to delete from the dictionary.
        """
        pieces: CommandArgList = [name, *terms]

        return await self.execute_module_command(
            CommandName.FT_DICTDEL, *pieces, callback=IntCallback()
        )

    @module_command(
        CommandName.FT_DICTDUMP,
        module=MODULE,
        version_introduced="1.4.0",
        group=COMMAND_GROUP,
    )
    async def dictdump(self, name: StringT) -> Set[AnyStr]:
        """
        Dumps all terms in the given dictionary

        :param name: The name of the dictionary to dump.
        """

        return await self.execute_module_command(
            CommandName.FT_DICTDUMP, name, callback=SetCallback[AnyStr]()
        )

    @module_command(
        CommandName.FT__LIST,
        module=MODULE,
        version_introduced="2.0.0",
        group=COMMAND_GROUP,
        cluster=ClusterCommandConfig(
            route=NodeFlag.PRIMARIES,
            combine=ClusterMergeSets[AnyStr](),
        ),
    )
    async def list(self) -> Set[AnyStr]:
        """
        Returns a list of all existing indexes
        """
        return await self.execute_module_command(
            CommandName.FT__LIST, callback=SetCallback[AnyStr]()
        )

    @module_command(
        CommandName.FT_CONFIG_SET,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
        cluster=ClusterCommandConfig(
            route=NodeFlag.PRIMARIES,
            combine=ClusterEnsureConsistent[bool](),
        ),
    )
    async def config_set(self, option: StringT, value: ValueT) -> bool:
        """
        Sets runtime configuration options
        """

        return await self.execute_module_command(
            CommandName.FT_CONFIG_SET, option, value, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.FT_CONFIG_GET,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
        cluster=ClusterCommandConfig(
            route=NodeFlag.RANDOM,
        ),
    )
    async def config_get(self, option: StringT) -> Dict[AnyStr, ResponsePrimitive]:
        """
        Retrieves runtime configuration options
        """

        return await self.execute_module_command(
            CommandName.FT_CONFIG_GET,
            option,
            callback=SearchConfigCallback[AnyStr](),
        )

    @module_command(
        CommandName.FT_SEARCH,
        module=MODULE,
        version_introduced="1.0.0",
        group=COMMAND_GROUP,
        arguments={"dialect": {"version_introduced": "2.4.3"}},
    )
    async def search(
        self,
        index: KeyT,
        query: StringT,
        *,
        nocontent: Optional[bool] = None,
        verbatim: Optional[bool] = None,
        nostopwords: Optional[bool] = None,
        withscores: Optional[bool] = None,
        withpayloads: Optional[bool] = None,
        withsortkeys: Optional[bool] = None,
        numeric_filters: Optional[
            Mapping[
                StringT, Tuple[Union[int, float, StringT], Union[int, float, StringT]]
            ]
        ] = None,
        geo_filters: Optional[
            Mapping[
                StringT,
                Tuple[
                    Tuple[Union[int, float], Union[int, float]],
                    Union[int, float],
                    Literal[PureToken.KM, PureToken.M, PureToken.MI, PureToken.FT],
                ],
            ]
        ] = None,
        in_keys: Optional[Parameters[StringT]] = None,
        in_fields: Optional[Parameters[StringT]] = None,
        returns: Optional[Mapping[StringT, Optional[StringT]]] = None,
        summarize_fields: Optional[Parameters[StringT]] = None,
        summarize_frags: Optional[int] = None,
        summarize_length: Optional[int] = None,
        summarize_separator: Optional[StringT] = None,
        highlight_fields: Optional[Parameters[StringT]] = None,
        highlight_tags: Optional[Tuple[StringT, StringT]] = None,
        slop: Optional[int] = None,
        timeout: Optional[Union[int, timedelta]] = None,
        inorder: Optional[bool] = None,
        language: Optional[StringT] = None,
        expander: Optional[StringT] = None,
        scorer: Optional[StringT] = None,
        explainscore: Optional[bool] = None,
        payload: Optional[StringT] = None,
        sortby: Optional[StringT] = None,
        sort_order: Optional[Literal[PureToken.ASC, PureToken.DESC]] = None,
        offset: Optional[int] = 0,
        limit: Optional[int] = None,
        parameters: Optional[Mapping[StringT, ValueT]] = None,
        dialect: Optional[int] = None,
    ) -> SearchResult[AnyStr]:
        """
        Searches the index with a textual query, returning either documents or just ids

        :param index: The name of the index to search.
        :param query: The text query to search.
        :param nocontent: If ``True``, returns the document ids and not the content.
        :param verbatim: If ``True``, does not try to use stemming for query expansion
         but searches the query terms verbatim.
        :param nostopwords: If ``True``, disables stopword filtering.
        :param withscores: If ``True``, also returns the relative internal score of each document.
        :param withpayloads: If ``True``, retrieves optional document payloads.
        :param withsortkeys: If ``True``, returns the value of the sorting key
        :param numeric_filters: A dictionary of numeric filters.
         Limits results to those having numeric values ranging between min and max,
        :param geo_filters: A dictionary of geo filters.
         Filters the results to a given radius from lon and lat.
         Radius is given as a number and units.
        :param in_keys: A list of keys to limit the result to.
         Non-existent keys are ignored, unless all the keys are non-existent.
        :param in_fields: Filters the results to those appearing only in specific
         attributes of the document, like title or URL.
        :param returns: A dictionary of attributes to return from the document. The values
         in the dictionary are used as an alias for the attribute name and if ``None``,
         the attribute will be returned as is.
        :param summarize_fields: A list of fields to summarize. If not ``None``,
         returns a summary of the fields.
        :param summarize_frags: The number of fragments to return in the summary.
        :param summarize_length: The length of each fragment in the summary.
        :param summarize_separator: The separator to use between fragments in the summary.
        :param highlight_fields: A list of fields to highlight.
        :param highlight_tags: A tuple of opening and closing tags to use for highlighting.
        :param slop: The number of words allowed between query terms.
        :param timeout: The timeout for the search query.
        :param inorder: If ``True``, requires that the query terms appear in the same order
         as in the query.
        :param language: The language to use for stemming.
        :param expander: The query expander to use.
        :param scorer: The scorer to use.
        :param explainscore: If ``True``, returns an explanation of the score.
        :param payload: The payload to return.
        :param sortby: The field to sort by.
        :param sort_order: The order to sort by.
        :param offset: The offset to start returning results from.
        :param limit: The maximum number of results to return.
        :param parameters: A dictionary of parameters to pass to the query.
        :param dialect: The query dialect to use.

        """
        pieces: CommandArgList = [index, query]
        if nocontent:
            pieces.append(PureToken.NOCONTENT)
        if verbatim:
            pieces.append(PureToken.VERBATIM)
        if nostopwords:
            pieces.append(PureToken.NOSTOPWORDS)
        if withscores:
            pieces.append(PureToken.WITHSCORES)
        if withpayloads:
            pieces.append(PureToken.WITHPAYLOADS)
        if withsortkeys:
            pieces.append(PureToken.WITHSORTKEYS)
        if numeric_filters:
            for field, numeric_filter in numeric_filters.items():
                pieces.extend(
                    [PrefixToken.FILTER, field, numeric_filter[0], numeric_filter[1]]
                )
        if geo_filters:
            for field, gfilter in geo_filters.items():
                pieces.extend(
                    [
                        PrefixToken.GEOFILTER,
                        field,
                        gfilter[0][0],
                        gfilter[0][1],
                        gfilter[1],
                        gfilter[2],
                    ]
                )
        if in_keys:
            _in_keys: List[StringT] = list(in_keys)
            pieces.extend([PrefixToken.INKEYS, len(_in_keys), *_in_keys])
        if in_fields:
            _in_fields: List[StringT] = list(in_fields)
            pieces.extend([PrefixToken.INFIELDS, len(_in_fields), *_in_fields])
        if returns:
            return_items: CommandArgList = []
            for identifier, property in returns.items():
                return_items.append(identifier)
                if property:
                    return_items.extend([PrefixToken.AS, property])
            pieces.extend([PrefixToken.RETURN, len(return_items), *return_items])
        if sortby:
            pieces.extend([PrefixToken.SORTBY, sortby])
            if sort_order:
                pieces.append(sort_order)
        if (
            summarize_fields
            or summarize_frags
            or summarize_length
            or summarize_separator
        ):
            pieces.append(PureToken.SUMMARIZE)
            if summarize_fields:
                _fields: List[StringT] = list(summarize_fields)
                pieces.extend([PrefixToken.FIELDS, len(_fields), *_fields])
            if summarize_frags:
                pieces.extend([PrefixToken.FRAGS, summarize_frags])
            if summarize_length:
                pieces.extend([PrefixToken.LEN, summarize_length])
            if summarize_separator:
                pieces.extend([PrefixToken.SEPARATOR, summarize_separator])
        if highlight_fields or highlight_tags:
            pieces.append(PureToken.HIGHLIGHT)
            if highlight_fields:
                _fields = list(highlight_fields)
                pieces.extend([PrefixToken.FIELDS, len(_fields), *_fields])
            if highlight_tags:
                pieces.extend([PureToken.TAGS, highlight_tags[0], highlight_tags[1]])
        if slop is not None:
            pieces.extend([PrefixToken.SLOP, slop])
        if timeout:
            pieces.extend([PrefixToken.TIMEOUT, normalized_milliseconds(timeout)])
        if inorder:
            pieces.append(PureToken.INORDER)
        if language:
            pieces.extend([PrefixToken.LANGUAGE, language])
        if expander:  # noqa
            pieces.extend([PrefixToken.EXPANDER, expander])
        if scorer:  # noqa
            pieces.extend([PrefixToken.SCORER, scorer])
        if explainscore:
            pieces.append(PureToken.EXPLAINSCORE)
        if payload:
            pieces.extend([PrefixToken.PAYLOAD, payload])
        if limit is not None:
            pieces.extend([PrefixToken.LIMIT, offset or 0, limit])
        if parameters:
            _parameters: List[ValueT] = list(itertools.chain(*parameters.items()))
            pieces.extend([PureToken.PARAMS, len(_parameters), *_parameters])
        if dialect:
            pieces.extend([PrefixToken.DIALECT, dialect])

        return await self.execute_module_command(
            CommandName.FT_SEARCH,
            *pieces,
            callback=SearchResultCallback[AnyStr](),
            withscores=withscores,
            withpayloads=withpayloads,
            withsortkeys=withsortkeys,
            explainscore=explainscore,
            nocontent=nocontent,
        )

    @module_command(
        CommandName.FT_AGGREGATE,
        module=MODULE,
        version_introduced="1.1.0",
        group=COMMAND_GROUP,
        arguments={"dialect": {"version_introduced": "2.4.3"}},
    )
    async def aggregate(
        self,
        index: KeyT,
        query: StringT,
        *,
        verbatim: Optional[bool] = None,
        load: Optional[
            Union[Literal["*"], Parameters[Union[StringT, Tuple[StringT, StringT]]]]
        ] = None,
        timeout: Optional[Union[int, timedelta]] = None,
        transforms: Optional[Parameters[Union[Group, Apply, Filter]]] = None,
        sortby: Optional[
            Mapping[StringT, Literal[PureToken.ASC, PureToken.DESC]]
        ] = None,
        sortby_max: Optional[int] = None,
        offset: Optional[int] = 0,
        limit: Optional[int] = None,
        with_cursor: Optional[bool] = None,
        cursor_read_size: Optional[int] = None,
        cursor_maxidle: Optional[Union[int, timedelta]] = None,
        parameters: Optional[Mapping[StringT, StringT]] = None,
        dialect: Optional[int] = None,
    ) -> SearchAggregationResult[AnyStr]:
        """
        Perform aggregate transformations on search results from a Redis index.

        :param index: Name of the Redis index to search.
        :param query: Base filtering query to retrieve documents.
        :param verbatim: If ``True``, search query terms verbatim.
        :param load: Load document attributes from the source document.
        :param timeout: Maximum time to wait for the query to complete.
        :param transforms: List of transformations to apply to the results.
        :param sortby: Sort the pipeline up to the point of SORTBY.
        :param sortby_max: Optimize sorting by sorting only for the n-largest elements.
        :param offset: Number of results to skip.
        :param limit: Maximum number of results to return.
        :param with_cursor: If ``True``, return a cursor for large result sets.
        :param cursor_read_size: Number of results to read from the cursor at a time.
        :param cursor_maxidle: Maximum idle time for the cursor.
        :param parameters: Additional parameters to pass to the query.
        :param dialect: Query dialect to use.

        :return: Aggregated search results from the Redis index.

        """
        pieces: CommandArgList = [index, query]
        if verbatim:
            pieces.append(PureToken.VERBATIM)
        if timeout:
            pieces.extend([PrefixToken.TIMEOUT, normalized_milliseconds(timeout)])
        if load:
            pieces.append(PrefixToken.LOAD)
            if isinstance(load, (bytes, str)):
                pieces.append(load)
            else:
                _load_fields: List[StringT] = []
                for field in load:
                    if isinstance(field, (bytes, str)):
                        _load_fields.append(field)
                    else:
                        _load_fields.extend([field[0], PrefixToken.AS, field[1]])

                pieces.extend([len(_load_fields), *_load_fields])

        if transforms:
            for step in transforms:
                pieces.extend(step.args)

        if sortby:
            pieces.append(PrefixToken.SORTBY)
            pieces.append(len(sortby) * 2)
            for field, order in sortby.items():
                pieces.extend([field, order])
            if sortby_max:
                pieces.extend([PrefixToken.MAX, sortby_max])

        if limit is not None:
            pieces.extend([PrefixToken.LIMIT, offset or 0, limit])

        if with_cursor:
            pieces.append(PureToken.WITHCURSOR)
            if cursor_read_size:
                pieces.extend([PrefixToken.COUNT, cursor_read_size])
            if cursor_maxidle:
                pieces.extend(
                    [PrefixToken.MAXIDLE, normalized_milliseconds(cursor_maxidle)]
                )
        if parameters:
            _parameters: List[StringT] = list(itertools.chain(*parameters.items()))
            pieces.extend([PureToken.PARAMS, len(_parameters), *_parameters])
        if dialect:
            pieces.extend([PrefixToken.DIALECT, dialect])
        return await self.execute_module_command(
            CommandName.FT_AGGREGATE,
            *pieces,
            callback=AggregationResultCallback[AnyStr](),
            with_cursor=with_cursor,
            dialect=dialect,
        )

    @module_command(
        CommandName.FT_CURSOR_READ,
        module=MODULE,
        version_introduced="1.1.0",
        group=COMMAND_GROUP,
    )
    async def cursor_read(
        self, index: KeyT, cursor_id: int, count: Optional[int] = None
    ) -> SearchAggregationResult[AnyStr]:
        """
        Reads from a cursor
        """
        pieces: CommandArgList = [index, cursor_id]
        if count:
            pieces.extend([PrefixToken.COUNT, count])

        return await self.execute_module_command(
            CommandName.FT_CURSOR_READ,
            *pieces,
            callback=AggregationResultCallback[AnyStr](),
            with_cursor=True,
        )

    @module_command(
        CommandName.FT_CURSOR_DEL,
        module=MODULE,
        version_introduced="1.1.0",
        group=COMMAND_GROUP,
    )
    async def cursor_del(self, index: KeyT, cursor_id: int) -> bool:
        """
        Deletes a cursor

        """
        pieces: CommandArgList = [index, cursor_id]

        return await self.execute_module_command(
            CommandName.FT_CURSOR_DEL, *pieces, callback=SimpleStringCallback()
        )



from __future__ import annotations

import enum

from coredis._protocols import AbstractExecutor
from coredis.commands.constants import CommandName
from coredis.exceptions import ReadOnlyError
from coredis.tokens import PrefixToken, PureToken
from coredis.typing import (
    AnyStr,
    CommandArgList,
    Generic,
    KeyT,
    Literal,
    ResponseType,
    Union,
)


class BitFieldSubCommand(bytes, enum.Enum):
    SET = PrefixToken.SET
    GET = PrefixToken.GET
    INCRBY = PrefixToken.INCRBY
    OVERFLOW = PrefixToken.OVERFLOW


class BitFieldOperation(Generic[AnyStr]):
    """
    The command treats a Redis string as a array of bits,
    and is capable of addressing specific integer fields
    of varying bit widths and arbitrary non (necessary) aligned offset.

    The supported types are up to 64 bits for signed integers,
    and up to 63 bits for unsigned integers.

    Offset can be num prefixed with `#` character or num directly.

    Redis command documentation: `BITFIELD <https://redios.io/commands/bitfield>`__
    """

    def __init__(
        self, redis_client: AbstractExecutor, key: KeyT, readonly: bool = False
    ) -> None:
        self._command = (
            CommandName.BITFIELD if not readonly else CommandName.BITFIELD_RO
        )
        self._command_stack: CommandArgList = [key]
        self.redis = redis_client
        self.readonly = readonly

    def __del__(self) -> None:
        self._command_stack.clear()

    def set(
        self, encoding: str, offset: Union[int, str], value: int
    ) -> BitFieldOperation[AnyStr]:
        """
        Set the specified bit field and returns its old value.
        """

        if self.readonly:
            raise ReadOnlyError()

        self._command_stack.extend([BitFieldSubCommand.SET, encoding, offset, value])

        return self

    def get(self, encoding: str, offset: Union[int, str]) -> BitFieldOperation[AnyStr]:
        """
        Returns the specified bit field.
        """

        self._command_stack.extend([BitFieldSubCommand.GET, encoding, offset])

        return self

    def incrby(
        self, encoding: str, offset: Union[int, str], increment: int
    ) -> BitFieldOperation[AnyStr]:
        """
        Increments or decrements (if a negative increment is given)
        the specified bit field and returns the new value.
        """

        if self.readonly:
            raise ReadOnlyError()

        self._command_stack.extend(
            [BitFieldSubCommand.INCRBY, encoding, offset, increment]
        )

        return self

    def overflow(
        self,
        behavior: Literal[
            PureToken.SAT, PureToken.WRAP, PureToken.FAIL
        ] = PureToken.SAT,
    ) -> BitFieldOperation[AnyStr]:
        """
        fine-tune the behavior of the increment or decrement overflow,
        have no effect unless used before :meth:`incrby`
        three :paramref:`behavior` types are available: ``WRAP|SAT|FAIL``
        """

        if self.readonly:
            raise ReadOnlyError()
        self._command_stack.extend([BitFieldSubCommand.OVERFLOW, behavior])

        return self

    async def exc(self) -> ResponseType:
        """execute commands in command stack"""

        return await self.redis.execute_command(
            self._command, *self._command_stack, decode=False
        )



from __future__ import annotations

from coredis._utils import EncodingInsensitiveDict
from coredis.response._callbacks import ResponseCallback, SimpleStringCallback
from coredis.response.types import LCSMatch, LCSResult
from coredis.typing import (
    AnyStr,
    Dict,
    List,
    Optional,
    ResponsePrimitive,
    ResponseType,
    Union,
    ValueT,
)


class StringSetCallback(
    ResponseCallback[Optional[AnyStr], Optional[AnyStr], Optional[Union[AnyStr, bool]]]
):
    def transform(
        self, response: Optional[AnyStr], **options: Optional[ValueT]
    ) -> Optional[Union[AnyStr, bool]]:
        if options.get("get"):
            return response
        else:
            return SimpleStringCallback()(response)


class LCSCallback(
    ResponseCallback[
        List[ResponseType],
        Dict[ResponsePrimitive, ResponseType],
        Union[AnyStr, int, LCSResult],
    ]
):
    def transform(
        self,
        response: Union[
            List[ResponseType],
            Dict[ResponsePrimitive, ResponseType],
        ],
        **options: Optional[ValueT],
    ) -> LCSResult:
        assert (
            isinstance(response, list)
            and isinstance(response[-1], int)
            and isinstance(response[1], list)
        )
        return LCSResult(
            tuple(
                LCSMatch(
                    (int(k[0][0]), int(k[0][1])),
                    (int(k[1][0]), int(k[1][1])),
                    k[2] if len(k) > 2 else None,
                )
                for k in response[1]
            ),
            response[-1],
        )

    def transform_3(
        self,
        response: Dict[ResponsePrimitive, ResponseType],
        **options: Optional[ValueT],
    ) -> LCSResult:
        proxy = EncodingInsensitiveDict(response)
        return LCSResult(
            tuple(
                LCSMatch(
                    (int(k[0][0]), int(k[0][1])),
                    (int(k[1][0]), int(k[1][1])),
                    k[2] if len(k) > 2 else None,
                )
                for k in proxy["matches"]
            ),
            proxy["len"],
        )



from __future__ import annotations

import asyncio
import logging
from abc import ABC, abstractmethod
from functools import wraps
from typing import Any

from coredis.typing import Callable, Coroutine, Dict, Optional, P, R, Tuple, Type, Union

logger = logging.getLogger(__name__)


class RetryPolicy(ABC):
    """
    Abstract retry policy
    """

    def __init__(
        self, retries: int, retryable_exceptions: Tuple[Type[BaseException], ...]
    ) -> None:
        """
        :param retries: number of times to retry if a :paramref:`retryable_exception`
         is encountered.
        :param retryable_exceptions: The exceptions to trigger a retry for
        """
        self.retryable_exceptions = retryable_exceptions
        self.retries = retries

    @abstractmethod
    async def delay(self, attempt_number: int) -> None:
        pass

    async def call_with_retries(
        self,
        func: Callable[..., Coroutine[Any, Any, R]],
        before_hook: Optional[Callable[..., Coroutine[Any, Any, Any]]] = None,
        failure_hook: Optional[
            Union[
                Callable[..., Coroutine[Any, Any, None]],
                Dict[Type[BaseException], Callable[..., Coroutine[Any, Any, None]]],
            ]
        ] = None,
    ) -> R:
        """
        :param func: a function that should return the coroutine that will be
         awaited when retrying if :paramref:`RetryPolicy.retryable_exceptions` is encountered.
        :param before_hook: if provided will be called on every attempt.
        :param failure_hook: if provided and is a callable it will be
         called everytime a retryable exception is encountered. If it is a mapping
         of exception types to callables, the first exception type that is a parent
         of any encountered exception will be called.
        """
        last_error: Optional[BaseException] = None
        for attempt in range(self.retries + 1):
            try:
                await self.delay(attempt)
                if before_hook:
                    await before_hook()
                return await func()
            except self.retryable_exceptions as e:
                logger.info(f"Retry attempt {attempt + 1} due to error: {e}")
                if failure_hook:
                    try:
                        if isinstance(failure_hook, dict):
                            for exc_type, hook in failure_hook.items():
                                if isinstance(e, exc_type):
                                    await hook(e)
                                    break
                        else:
                            await failure_hook(e)
                    except:  # noqa
                        pass
                last_error = e
        if last_error:
            raise last_error
        assert False

    def will_retry(self, exc: BaseException) -> bool:
        return isinstance(exc, self.retryable_exceptions)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}<"
            f"retries={self.retries}, "
            f"retryable_exceptions={','.join(e.__name__ for e in self.retryable_exceptions)}"
            ">"
        )


class NoRetryPolicy(RetryPolicy):
    def __init__(self) -> None:
        super().__init__(1, ())

    async def delay(self, attempt_number: int) -> None:
        pass


class ConstantRetryPolicy(RetryPolicy):
    """
    Retry policy that pauses :paramref:`ConstantRetryPolicy.delay`
    seconds between :paramref:`ConstantRetryPolicy.retries`
    if any of :paramref:`ConstantRetryPolicy.retryable_exceptions` are
    encountered.
    """

    def __init__(
        self,
        retryable_exceptions: Tuple[Type[BaseException], ...],
        retries: int,
        delay: float,
    ) -> None:
        self.__delay = delay
        super().__init__(retries, retryable_exceptions)

    async def delay(self, attempt_number: int) -> None:
        if attempt_number > 0:
            await asyncio.sleep(self.__delay)


class ExponentialBackoffRetryPolicy(RetryPolicy):
    """
    Retry policy that exponentially backs off before retrying up to
    :paramref:`ExponentialBackoffRetryPolicy.retries` if any
    of :paramref:`ExponentialBackoffRetryPolicy.retryable_exceptions` are
    encountered. :paramref:`ExponentialBckoffRetryPolicy.initial_delay`
    is used as the initial value for calculating the exponential backoff.

    """

    def __init__(
        self,
        retryable_exceptions: Tuple[Type[BaseException], ...],
        retries: int,
        initial_delay: float,
    ) -> None:
        self.__initial_delay = initial_delay
        super().__init__(retries, retryable_exceptions)

    async def delay(self, attempt_number: int) -> None:
        if attempt_number > 0:
            await asyncio.sleep(pow(2, attempt_number) * self.__initial_delay)


class CompositeRetryPolicy(RetryPolicy):
    """
    Convenience class to combine multiple retry policies
    """

    def __init__(self, *retry_policies: RetryPolicy):
        self._retry_policies = set(retry_policies)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{','.join(str(p) for p in self._retry_policies)}>"

    def add_retry_policy(self, policy: RetryPolicy) -> None:
        """
        Add to the retry policies that this instance was created with
        """
        self._retry_policies.add(policy)

    async def delay(self, attempt_number: int) -> None:
        raise NotImplementedError()

    async def call_with_retries(
        self,
        func: Callable[..., Coroutine[Any, Any, R]],
        before_hook: Optional[Callable[..., Coroutine[Any, Any, Any]]] = None,
        failure_hook: Optional[
            Union[
                Callable[..., Coroutine[Any, Any, None]],
                Dict[Type[BaseException], Callable[..., Coroutine[Any, Any, None]]],
            ]
        ] = None,
    ) -> R:
        """
        Calls :paramref:`func` repeatedly according to the retry policies that
        this class was instantiated with (:paramref:`CompositeRetryPolicy.retry_policies`).

        :param func: a function that should return the coroutine that will be
         awaited when retrying if :paramref:`RetryPolicy.retryable_exceptions` is encountered.
        :param before_hook: if provided will be called before every attempt.
        :param failure_hook: if provided and is a callable it will be
         called everytime a retryable exception is encountered. If it is a mapping
         of exception types to callables, the first exception type that is a parent
         of any encountered exception will be called.
        """
        attempts = {policy: 0 for policy in self._retry_policies}
        while True:
            try:
                if before_hook:
                    await before_hook()
                return await func()
            except Exception as e:
                will_retry = False
                attempt = 0
                for policy in attempts:
                    if policy.will_retry(e) and attempts[policy] < policy.retries:
                        attempt = attempts[policy]
                        attempts[policy] += 1
                        await policy.delay(attempts[policy])
                        will_retry = True
                        break

                if failure_hook:
                    if isinstance(failure_hook, dict):
                        for exc_type in failure_hook:
                            if isinstance(e, exc_type):
                                await failure_hook[exc_type](e)
                                break
                    else:
                        await failure_hook(e)

                if will_retry:
                    logger.info(f"Retry attempt {attempt} due to error: {e}")
                    continue

                raise e


def retryable(
    policy: RetryPolicy,
    failure_hook: Optional[Callable[..., Coroutine[Any, Any, None]]] = None,
) -> Callable[
    [Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]
]:
    """
    Decorator to be used to apply a retry policy to a coroutine
    """

    def inner(
        func: Callable[P, Coroutine[Any, Any, R]]
    ) -> Callable[P, Coroutine[Any, Any, R]]:
        @wraps(func)
        async def _inner(*args: P.args, **kwargs: P.kwargs) -> R:
            return await policy.call_with_retries(
                lambda: func(*args, **kwargs), failure_hook=failure_hook
            )

        return _inner

    return inner



from __future__ import annotations

import asyncio
import contextlib
import contextvars
import functools
import warnings
from collections import defaultdict
from ssl import SSLContext
from typing import TYPE_CHECKING, Any, cast, overload

from deprecated.sphinx import versionadded
from packaging import version
from packaging.version import InvalidVersion, Version

from coredis._utils import EncodingInsensitiveDict, nativestr
from coredis.cache import AbstractCache, SupportsClientTracking
from coredis.commands._key_spec import KeySpec
from coredis.commands.constants import CommandFlag, CommandName
from coredis.commands.core import CoreCommands
from coredis.commands.function import Library
from coredis.commands.monitor import Monitor
from coredis.commands.pubsub import PubSub
from coredis.commands.script import Script
from coredis.commands.sentinel import SentinelCommands
from coredis.config import Config
from coredis.connection import (
    BaseConnection,
    RedisSSLContext,
    UnixDomainSocketConnection,
)
from coredis.exceptions import (
    ConnectionError,
    PersistenceError,
    RedisError,
    ReplicationError,
    TimeoutError,
    UnknownCommandError,
    WatchError,
)
from coredis.globals import COMMAND_FLAGS, READONLY_COMMANDS
from coredis.modules import ModuleMixin
from coredis.pool import ConnectionPool
from coredis.response._callbacks import (
    AsyncPreProcessingCallback,
    NoopCallback,
    ResponseCallback,
)
from coredis.response.types import ScoredMember
from coredis.retry import ConstantRetryPolicy, NoRetryPolicy, RetryPolicy
from coredis.typing import (
    AnyStr,
    AsyncGenerator,
    AsyncIterator,
    Callable,
    ContextManager,
    Coroutine,
    Dict,
    Generator,
    Generic,
    Iterator,
    KeyT,
    Literal,
    Optional,
    Parameters,
    ParamSpec,
    ResponseType,
    StringT,
    Tuple,
    Type,
    TypeVar,
    ValueT,
)

P = ParamSpec("P")
R = TypeVar("R")

if TYPE_CHECKING:
    import coredis.pipeline


ClientT = TypeVar("ClientT", bound="Client[Any]")
RedisT = TypeVar("RedisT", bound="Redis[Any]")
RedisStringT = TypeVar("RedisStringT", bound="Redis[str]")
RedisBytesT = TypeVar("RedisBytesT", bound="Redis[bytes]")


class Client(
    Generic[AnyStr],
    CoreCommands[AnyStr],
    ModuleMixin[AnyStr],
    SentinelCommands[AnyStr],
):
    cache: Optional[AbstractCache]
    connection_pool: ConnectionPool
    decode_responses: bool
    encoding: str
    protocol_version: Literal[2, 3]
    server_version: Optional[Version]
    callback_storage: Dict[Type[ResponseCallback[Any, Any, Any]], Dict[str, Any]]

    def __init__(
        self,
        host: Optional[str] = "localhost",
        port: Optional[int] = 6379,
        db: int = 0,
        username: Optional[str] = None,
        password: Optional[str] = None,
        stream_timeout: Optional[float] = None,
        connect_timeout: Optional[float] = None,
        connection_pool: Optional[ConnectionPool] = None,
        connection_pool_cls: Type[ConnectionPool] = ConnectionPool,
        unix_socket_path: Optional[str] = None,
        encoding: str = "utf-8",
        decode_responses: bool = False,
        ssl: bool = False,
        ssl_context: Optional[SSLContext] = None,
        ssl_keyfile: Optional[str] = None,
        ssl_certfile: Optional[str] = None,
        ssl_cert_reqs: Optional[Literal["optional", "required", "none"]] = None,
        ssl_check_hostname: Optional[bool] = None,
        ssl_ca_certs: Optional[str] = None,
        max_connections: Optional[int] = None,
        max_idle_time: float = 0,
        idle_check_interval: float = 1,
        client_name: Optional[str] = None,
        protocol_version: Literal[2, 3] = 3,
        verify_version: bool = True,
        noreply: bool = False,
        retry_policy: RetryPolicy = NoRetryPolicy(),
        noevict: bool = False,
        notouch: bool = False,
        **kwargs: Any,
    ):
        if not connection_pool:
            kwargs = {
                "db": db,
                "username": username,
                "password": password,
                "encoding": encoding,
                "stream_timeout": stream_timeout,
                "connect_timeout": connect_timeout,
                "max_connections": max_connections,
                "decode_responses": decode_responses,
                "max_idle_time": max_idle_time,
                "idle_check_interval": idle_check_interval,
                "client_name": client_name,
                "protocol_version": protocol_version,
                "noreply": noreply,
                "noevict": noevict,
                "notouch": notouch,
            }

            if unix_socket_path is not None:
                kwargs.update(
                    {
                        "path": unix_socket_path,
                        "connection_class": UnixDomainSocketConnection,
                    }
                )
            else:
                # TCP specific options
                kwargs.update({"host": host, "port": port})

                if ssl_context is not None:
                    kwargs["ssl_context"] = ssl_context
                elif ssl:
                    ssl_context = RedisSSLContext(
                        ssl_keyfile,
                        ssl_certfile,
                        ssl_cert_reqs,
                        ssl_ca_certs,
                        ssl_check_hostname,
                    ).get()
                    kwargs["ssl_context"] = ssl_context
            connection_pool = connection_pool_cls(**kwargs)

        self.connection_pool = connection_pool
        self.encoding = str(connection_pool.connection_kwargs.get("encoding", encoding))
        self.decode_responses = bool(
            connection_pool.connection_kwargs.get("decode_responses", decode_responses)
        )
        connection_protocol_version = (
            connection_pool.connection_kwargs.get("protocol_version")
            or protocol_version
        )
        assert connection_protocol_version in {
            2,
            3,
        }, "Protocol version can only be one of {2,3}"
        self.protocol_version = connection_protocol_version
        self.server_version: Optional[Version] = None
        self.verify_version = verify_version
        self.__noreply = noreply
        self._noreplycontext: contextvars.ContextVar[Optional[bool]] = (
            contextvars.ContextVar("noreply", default=None)
        )
        self._waitcontext: contextvars.ContextVar[Optional[Tuple[int, int]]] = (
            contextvars.ContextVar("wait", default=None)
        )
        self._waitaof_context: contextvars.ContextVar[
            Optional[Tuple[int, int, int]]
        ] = contextvars.ContextVar("waitaof", default=None)
        self.retry_policy = retry_policy
        self._module_info: Optional[Dict[str, version.Version]] = None
        self.callback_storage = defaultdict(lambda: {})

    @property
    def noreply(self) -> bool:
        if not hasattr(self, "_noreplycontext"):
            return False
        ctx = self._noreplycontext.get()
        if ctx is not None:
            return ctx
        return self.__noreply

    @property
    def requires_wait(self) -> bool:
        if not hasattr(self, "_waitcontext") or not self._waitcontext.get():
            return False
        return True

    @property
    def requires_waitaof(self) -> bool:
        if not hasattr(self, "_waitaof_context") or not self._waitaof_context.get():
            return False
        return True

    async def get_server_module_version(self, module: str) -> Optional[version.Version]:
        if self._module_info is None:
            await self._populate_module_versions()
        return (self._module_info or {}).get(module)

    def _ensure_server_version(self, version: Optional[str]) -> None:
        if not self.verify_version or Config.optimized:
            return
        if not version:
            return
        if not self.server_version and version:
            try:
                self.server_version = Version(nativestr(version))
            except InvalidVersion:
                warnings.warn(
                    (
                        f"Server reported an invalid version: {version}."
                        "If this is expected you can dismiss this warning by passing "
                        "verify_version=False to the client constructor"
                    ),
                    category=UserWarning,
                )
                self.verify_version = False
                self.server_version = None

    async def _ensure_wait(
        self, command: bytes, connection: BaseConnection
    ) -> asyncio.Future[None]:
        maybe_wait: asyncio.Future[None] = asyncio.get_running_loop().create_future()
        wait = self._waitcontext.get()
        if wait and wait[0] > 0:

            def check_wait(
                wait: Tuple[int, int], response: asyncio.Future[ResponseType]
            ) -> None:
                exc = response.exception()
                if exc:
                    maybe_wait.set_exception(exc)
                elif not cast(int, response.result()) >= wait[0]:
                    maybe_wait.set_exception(
                        ReplicationError(command, wait[0], wait[1])
                    )
                else:
                    maybe_wait.set_result(None)

            request = await connection.create_request(
                CommandName.WAIT, *wait, decode=False
            )
            request.add_done_callback(functools.partial(check_wait, wait))
        else:
            maybe_wait.set_result(None)
        return maybe_wait

    async def _ensure_persistence(
        self, command: bytes, connection: BaseConnection
    ) -> asyncio.Future[None]:
        maybe_wait: asyncio.Future[None] = asyncio.get_running_loop().create_future()
        waitaof = self._waitaof_context.get()
        if waitaof and waitaof[0] > 0:

            def check_wait(
                waitaof: Tuple[int, int, int], response: asyncio.Future[ResponseType]
            ) -> None:
                exc = response.exception()
                if exc:
                    maybe_wait.set_exception(exc)
                else:
                    res = cast(Tuple[int, int], response.result())
                    if not (res[0] >= waitaof[0] and res[1] >= waitaof[1]):
                        maybe_wait.set_exception(PersistenceError(command, *waitaof))
                    else:
                        maybe_wait.set_result(None)

            request = await connection.create_request(
                CommandName.WAITAOF, *waitaof, decode=False
            )
            request.add_done_callback(functools.partial(check_wait, waitaof))
        else:
            maybe_wait.set_result(None)
        return maybe_wait

    async def _populate_module_versions(self) -> None:
        if self.noreply:
            return
        try:
            modules = await self.module_list()
            self._module_info = defaultdict(lambda: version.Version("0"))
            for module in modules:
                mod = EncodingInsensitiveDict(module)
                name = nativestr(mod["name"])
                ver = mod["ver"]
                ver, patch = divmod(ver, 100)
                ver, minor = divmod(ver, 100)
                ver, major = divmod(ver, 100)
                self._module_info[name] = version.Version(f"{major}.{minor}.{patch}")
        except UnknownCommandError:
            self._module_info = {}

    async def initialize(self: ClientT) -> ClientT:
        await self.connection_pool.initialize()
        return self

    def __await__(self: ClientT) -> Generator[Any, None, ClientT]:
        return self.initialize().__await__()

    def __repr__(self) -> str:
        return f"{type(self).__name__}<{repr(self.connection_pool)}>"

    async def scan_iter(
        self,
        match: Optional[StringT] = None,
        count: Optional[int] = None,
        type_: Optional[StringT] = None,
    ) -> AsyncIterator[AnyStr]:
        """
        Make an iterator using the SCAN command so that the client doesn't
        need to remember the cursor position.
        """
        cursor = None

        while cursor != 0:
            cursor, data = await self.scan(
                cursor=cursor, match=match, count=count, type_=type_
            )

            for item in data:
                yield item

    async def sscan_iter(
        self,
        key: KeyT,
        match: Optional[StringT] = None,
        count: Optional[int] = None,
    ) -> AsyncIterator[AnyStr]:
        """
        Make an iterator using the SSCAN command so that the client doesn't
        need to remember the cursor position.
        """
        cursor = None

        while cursor != 0:
            cursor, data = await self.sscan(
                key, cursor=cursor, match=match, count=count
            )

            for item in data:
                yield item

    async def hscan_iter(
        self,
        key: KeyT,
        match: Optional[StringT] = None,
        count: Optional[int] = None,
    ) -> AsyncGenerator[Tuple[AnyStr, AnyStr], None]:
        """
        Make an iterator using the HSCAN command so that the client doesn't
        need to remember the cursor position.
        """
        cursor = None

        while cursor != 0:
            cursor, data = await self.hscan(
                key, cursor=cursor, match=match, count=count
            )

            for item in data.items():
                yield item

    async def zscan_iter(
        self,
        key: KeyT,
        match: Optional[StringT] = None,
        count: Optional[int] = None,
    ) -> AsyncIterator[ScoredMember]:
        """
        Make an iterator using the ZSCAN command so that the client doesn't
        need to remember the cursor position.
        """
        cursor = None

        while cursor != 0:
            cursor, data = await self.zscan(
                key,
                cursor=cursor,
                match=match,
                count=count,
            )

            for item in data:
                yield item

    def register_script(self, script: ValueT) -> Script[AnyStr]:
        """
        Registers a Lua :paramref:`script`

        :return: A :class:`coredis.commands.script.Script` instance that is
         callable and hides the complexity of dealing with scripts, keys, and
         shas.
        """
        return Script[AnyStr](self, script)  # type: ignore

    @versionadded(version="3.1.0")
    async def register_library(
        self, name: StringT, code: StringT, replace: bool = False
    ) -> Library[AnyStr]:
        """
        Register a new library

        :param name: name of the library
        :param code: raw code for the library
        :param replace: Whether to replace the library when intializing. If ``False``
         an exception will be raised if the library was already loaded in the target
         redis instance.
        """
        return await Library[AnyStr](self, name=name, code=code, replace=replace)

    @versionadded(version="3.1.0")
    async def get_library(self, name: StringT) -> Library[AnyStr]:
        """
        Fetch a pre registered library

        :param name: name of the library
        """
        return await Library[AnyStr](self, name)

    @contextlib.contextmanager
    def ignore_replies(self: ClientT) -> Iterator[ClientT]:
        """
        Context manager to run commands without waiting for a reply.

        Example::

            client = coredis.Redis()
            with client.ignore_replies():
                assert None == await client.set("fubar", 1), "noreply"
            assert True == await client.set("fubar", 1), "reply"
        """
        self._noreplycontext.set(True)
        try:
            yield self
        finally:
            self._noreplycontext.set(None)

    @contextlib.contextmanager
    def ensure_replication(
        self: ClientT, replicas: int = 1, timeout_ms: int = 100
    ) -> Iterator[ClientT]:
        """
        Context manager to ensure that commands executed within the context
        are replicated to atleast :paramref:`replicas` within
        :paramref:`timeout_ms` milliseconds.

        Internally this uses `WAIT <https://redis.io/commands/wait>`_ after
        each command executed within the context

        :raises: :exc:`coredis.exceptions.ReplicationError`

        Example::

            client = coredis.RedisCluster("localhost", 7000)
            with client.ensure_replication(1, 20):
                await client.set("fubar", 1)

        """
        self._waitcontext.set((replicas, timeout_ms))
        try:
            yield self
        finally:
            self._waitcontext.set(None)

    @versionadded(version="4.12.0")
    @contextlib.contextmanager
    def ensure_persistence(
        self: ClientT,
        local: Literal[0, 1] = 0,
        replicas: int = 0,
        timeout_ms: int = 100,
    ) -> Iterator[ClientT]:
        """
        Context manager to ensure that commands executed within the context
        are synced to the AOF of a :paramref:`local` host and/or :paramref:`replicas`
        within :paramref:`timeout_ms` milliseconds.

        Internally this uses `WAITAOF <https://redis.io/commands/waitaof>`_ after
        each command executed within the context

        :raises: :exc:`coredis.exceptions.PersistenceError`

        Example for standalone client::

            client = coredis.Redis()
            with client.ensure_persistence(1, 0, 20):
                await client.set("fubar", 1)

        Example for cluster::

            client = coredis.RedisCluster("localhost", 7000)
            with client.ensure_persistence(1, 1, 20):
                await client.set("fubar", 1)

        """
        self._waitaof_context.set((local, replicas, timeout_ms))
        try:
            yield self
        finally:
            self._waitaof_context.set(None)

    def should_quick_release(self, command: bytes) -> bool:
        return CommandFlag.BLOCKING not in COMMAND_FLAGS[command]


class Redis(Client[AnyStr]):
    connection_pool: ConnectionPool

    @overload
    def __init__(
        self: Redis[bytes],
        host: Optional[str] = ...,
        port: Optional[int] = ...,
        db: int = ...,
        *,
        username: Optional[str] = ...,
        password: Optional[str] = ...,
        stream_timeout: Optional[float] = ...,
        connect_timeout: Optional[float] = ...,
        connection_pool: Optional[ConnectionPool] = ...,
        connection_pool_cls: Type[ConnectionPool] = ...,
        unix_socket_path: Optional[str] = ...,
        encoding: str = ...,
        decode_responses: Literal[False] = ...,
        ssl: bool = ...,
        ssl_context: Optional[SSLContext] = ...,
        ssl_keyfile: Optional[str] = ...,
        ssl_certfile: Optional[str] = ...,
        ssl_cert_reqs: Optional[Literal["optional", "required", "none"]] = ...,
        ssl_check_hostname: Optional[bool] = ...,
        ssl_ca_certs: Optional[str] = ...,
        max_connections: Optional[int] = ...,
        max_idle_time: float = ...,
        idle_check_interval: float = ...,
        client_name: Optional[str] = ...,
        protocol_version: Literal[2, 3] = ...,
        verify_version: bool = ...,
        cache: Optional[AbstractCache] = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        **kwargs: Any,
    ) -> None: ...

    @overload
    def __init__(
        self: Redis[str],
        host: Optional[str] = ...,
        port: Optional[int] = ...,
        db: int = ...,
        *,
        username: Optional[str] = ...,
        password: Optional[str] = ...,
        stream_timeout: Optional[float] = ...,
        connect_timeout: Optional[float] = ...,
        connection_pool: Optional[ConnectionPool] = ...,
        connection_pool_cls: Type[ConnectionPool] = ...,
        unix_socket_path: Optional[str] = ...,
        encoding: str = ...,
        decode_responses: Literal[True],
        ssl: bool = ...,
        ssl_context: Optional[SSLContext] = ...,
        ssl_keyfile: Optional[str] = ...,
        ssl_certfile: Optional[str] = ...,
        ssl_cert_reqs: Optional[Literal["optional", "required", "none"]] = ...,
        ssl_check_hostname: Optional[bool] = ...,
        ssl_ca_certs: Optional[str] = ...,
        max_connections: Optional[int] = ...,
        max_idle_time: float = ...,
        idle_check_interval: float = ...,
        client_name: Optional[str] = ...,
        protocol_version: Literal[2, 3] = ...,
        verify_version: bool = ...,
        cache: Optional[AbstractCache] = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        **kwargs: Any,
    ) -> None: ...

    def __init__(
        self,
        host: Optional[str] = "localhost",
        port: Optional[int] = 6379,
        db: int = 0,
        *,
        username: Optional[str] = None,
        password: Optional[str] = None,
        stream_timeout: Optional[float] = None,
        connect_timeout: Optional[float] = None,
        connection_pool: Optional[ConnectionPool] = None,
        connection_pool_cls: Type[ConnectionPool] = ConnectionPool,
        unix_socket_path: Optional[str] = None,
        encoding: str = "utf-8",
        decode_responses: bool = False,
        ssl: bool = False,
        ssl_context: Optional[SSLContext] = None,
        ssl_keyfile: Optional[str] = None,
        ssl_certfile: Optional[str] = None,
        ssl_cert_reqs: Optional[Literal["optional", "required", "none"]] = None,
        ssl_check_hostname: Optional[bool] = None,
        ssl_ca_certs: Optional[str] = None,
        max_connections: Optional[int] = None,
        max_idle_time: float = 0,
        idle_check_interval: float = 1,
        client_name: Optional[str] = None,
        protocol_version: Literal[2, 3] = 3,
        verify_version: bool = True,
        cache: Optional[AbstractCache] = None,
        noreply: bool = False,
        noevict: bool = False,
        notouch: bool = False,
        retry_policy: RetryPolicy = ConstantRetryPolicy(
            (ConnectionError, TimeoutError), 2, 0.01
        ),
        **kwargs: Any,
    ) -> None:
        """

        Changes
          - .. versionadded:: 4.12.0

            - :paramref:`retry_policy`
            - :paramref:`noevict`
            - :paramref:`notouch`
            - :meth:`Redis.ensure_persistence` context manager
            - Redis Module support

              - RedisJSON: :attr:`Redis.json`
              - RedisBloom:

                - BloomFilter: :attr:`Redis.bf`
                - CuckooFilter: :attr:`Redis.cf`
                - CountMinSketch: :attr:`Redis.cms`
                - TopK: :attr:`Redis.topk`
                - TDigest: :attr:`Redis.tdigest`
              - RedisTimeSeries: :attr:`Redis.timeseries`
              - RedisGraph: :attr:`Redis.graph`
              - RediSearch:

                - Search & Aggregation: :attr:`Redis.search`
                - Autocomplete: Added :attr:`Redis.autocomplete`

          - .. versionchanged:: 4.12.0

            - Removed :paramref:`retry_on_timeout` constructor argument. Use
              :paramref:`retry_policy` instead.

          - .. versionadded:: 4.3.0

            - Added :paramref:`connection_pool_cls`

          - .. versionchanged:: 4.0.0

            - :paramref:`non_atomic_cross_slot` defaults to ``True``
            - :paramref:`protocol_version`` defaults to ``3``

          - .. versionadded:: 3.11.0

            - Added :paramref:`noreply`

          - .. versionadded:: 3.9.0

            - If :paramref:`cache` is provided the client will check & populate
              the cache for read only commands and invalidate it for commands
              that could change the key(s) in the request.

          - .. versionchanged:: 3.5.0

            - The :paramref:`verify_version` parameter now defaults to ``True``

          - .. versionadded:: 3.1.0

            - The :paramref:`protocol_version` and :paramref:`verify_version`
              :parameters were added


        :param host: The hostname of the redis server
        :param port: The port at which th redis server is listening on
        :param db: database number to switch to upon connection
        :param username: Username for authenticating with the redis server
        :param password: Password for authenticating with the redis server
        :param stream_timeout: Timeout (seconds) when reading responses from the server
        :param connect_timeout: Timeout (seconds) for establishing a connection to the server
        :param connection_pool: The connection pool instance to use. If not provided
         a new pool will be assigned to this client.
        :param connection_pool_cls: The connection pool class to use when constructing
         a connection pool for this instance.
        :param unix_socket_path: Path to the UDS which the redis server
         is listening at
        :param encoding: The codec to use to encode strings transmitted to redis
         and decode responses with. (See :ref:`handbook/encoding:encoding/decoding`)
        :param decode_responses: If ``True`` string responses from the server
         will be decoded using :paramref:`encoding` before being returned.
         (See :ref:`handbook/encoding:encoding/decoding`)
        :param ssl: Whether to use an SSL connection
        :param ssl_context: If provided the :class:`ssl.SSLContext` will be used when
         establishing the connection. Otherwise either the default context (if no other
         ssl related parameters are provided) or a custom context based on the other
         ``ssl_*`` parameters will be used.
        :param ssl_keyfile: Path of the private key to use
        :param ssl_certfile: Path to the certificate corresponding to :paramref:`ssl_keyfile`
        :param ssl_cert_reqs: Whether to try to verify the server's certificates and
         how to behave if verification fails (See :attr:`ssl.SSLContext.verify_mode`).
        :param ssl_check_hostname: Whether to enable hostname checking when establishing
         an ssl connection.
        :param ssl_ca_certs: Path to a concatenated certificate authority file or a directory
         containing several CA certifcates to use  for validating the server's certificates
         when :paramref:`ssl_cert_reqs` is not ``"none"``
         (See :meth:`ssl.SSLContext.load_verify_locations`).
        :param max_connections: Maximum capacity of the connection pool (Ignored if
         :paramref:`connection_pool` is not ``None``.
        :param max_idle_time: Maximum number of a seconds an unused connection is cached
         before it is disconnected.
        :param idle_check_interval: Periodicity of idle checks (seconds) to release idle
         connections.
        :param client_name: The client name to identifiy with the redis server
        :param protocol_version: Whether to use the RESP (``2``) or RESP3 (``3``)
         protocol for parsing responses from the server (Default ``3``).
         (See :ref:`handbook/response:redis response`)
        :param verify_version: Validate redis server version against the documented
         version introduced before executing a command and raises a
         :exc:`CommandNotSupportedError` error if the required version is higher than
         the reported server version
        :param cache: If provided the cache will be used to avoid requests for read only
         commands if the client has already requested the data and it hasn't been invalidated.
         The cache is responsible for any mutations to the keys that happen outside of this client
        :param noreply: If ``True`` the client will not request a response for any
         commands sent to the server.
        :param noevict: Ensures that connections from the client will be excluded from the
         client eviction process even if we're above the configured client eviction threshold.
        :param notouch: Ensures that commands sent by the client will not alter the LRU/LFU of
         the keys they access.
        :param retry_policy: The retry policy to use when interacting with the redis server

        """
        super().__init__(
            host=host,
            port=port,
            db=db,
            username=username,
            password=password,
            stream_timeout=stream_timeout,
            connect_timeout=connect_timeout,
            connection_pool=connection_pool,
            connection_pool_cls=connection_pool_cls,
            unix_socket_path=unix_socket_path,
            encoding=encoding,
            decode_responses=decode_responses,
            ssl=ssl,
            ssl_context=ssl_context,
            ssl_keyfile=ssl_keyfile,
            ssl_certfile=ssl_certfile,
            ssl_cert_reqs=ssl_cert_reqs,
            ssh_check_hostname=ssl_check_hostname,
            ssl_ca_certs=ssl_ca_certs,
            max_connections=max_connections,
            max_idle_time=max_idle_time,
            idle_check_interval=idle_check_interval,
            client_name=client_name,
            protocol_version=protocol_version,
            verify_version=verify_version,
            noreply=noreply,
            noevict=noevict,
            notouch=notouch,
            retry_policy=retry_policy,
            **kwargs,
        )
        self.cache = cache
        self._decodecontext: contextvars.ContextVar[Optional[bool],] = (
            contextvars.ContextVar("decode", default=None)
        )
        self._encodingcontext: contextvars.ContextVar[Optional[str],] = (
            contextvars.ContextVar("decode", default=None)
        )

    @classmethod
    @overload
    def from_url(
        cls: Type[RedisBytesT],
        url: str,
        db: Optional[int] = ...,
        *,
        decode_responses: Literal[False] = ...,
        protocol_version: Literal[2, 3] = ...,
        verify_version: bool = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        **kwargs: Any,
    ) -> RedisBytesT: ...

    @classmethod
    @overload
    def from_url(
        cls: Type[RedisStringT],
        url: str,
        db: Optional[int] = ...,
        *,
        decode_responses: Literal[True],
        protocol_version: Literal[2, 3] = ...,
        verify_version: bool = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        **kwargs: Any,
    ) -> RedisStringT: ...

    @classmethod
    def from_url(
        cls: Type[RedisT],
        url: str,
        db: Optional[int] = None,
        *,
        decode_responses: bool = False,
        protocol_version: Literal[2, 3] = 3,
        verify_version: bool = True,
        noreply: bool = False,
        noevict: bool = False,
        notouch: bool = False,
        retry_policy: RetryPolicy = ConstantRetryPolicy(
            (ConnectionError, TimeoutError), 2, 0.01
        ),
        **kwargs: Any,
    ) -> RedisT:
        """
        Return a Redis client object configured from the given URL, which must
        use either the `redis:// scheme
        <http://www.iana.org/assignments/uri-schemes/prov/redis>`_ for RESP
        connections or the ``unix://`` scheme for Unix domain sockets.

        For example:

        - ``redis://[:password]@localhost:6379/0``
        - ``rediss://[:password]@localhost:6379/0``
        - ``unix://[:password]@/path/to/socket.sock?db=0``

        :paramref:`url` and :paramref:`kwargs` are passed as is to
        the :func:`coredis.ConnectionPool.from_url`.
        """
        if decode_responses:
            return cls(
                decode_responses=True,
                protocol_version=protocol_version,
                verify_version=verify_version,
                noreply=noreply,
                retry_policy=retry_policy,
                connection_pool=ConnectionPool.from_url(
                    url,
                    db=db,
                    decode_responses=decode_responses,
                    protocol_version=protocol_version,
                    noreply=noreply,
                    noevict=noevict,
                    notouch=notouch,
                    **kwargs,
                ),
            )
        else:
            return cls(
                decode_responses=False,
                protocol_version=protocol_version,
                verify_version=verify_version,
                noreply=noreply,
                retry_policy=retry_policy,
                connection_pool=ConnectionPool.from_url(
                    url,
                    db=db,
                    decode_responses=decode_responses,
                    protocol_version=protocol_version,
                    noreply=noreply,
                    noevict=noevict,
                    notouch=notouch,
                    **kwargs,
                ),
            )

    async def initialize(self) -> Redis[AnyStr]:
        if not self.connection_pool.initialized:
            await super().initialize()
            if self.cache:
                self.cache = await self.cache.initialize(self)
        return self

    async def execute_command(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., R] = NoopCallback(),
        **options: Optional[ValueT],
    ) -> R:
        """
        Executes a command with configured retries and returns
        the parsed response
        """
        return await self.retry_policy.call_with_retries(
            lambda: self._execute_command(command, *args, callback=callback, **options),
            before_hook=self.initialize,
        )

    async def _execute_command(
        self,
        command: bytes,
        *args: ValueT,
        callback: Callable[..., R] = NoopCallback(),
        **options: Optional[ValueT],
    ) -> R:
        pool = self.connection_pool
        quick_release = self.should_quick_release(command)
        connection = await pool.get_connection(
            command,
            *args,
            acquire=not quick_release or self.requires_wait or self.requires_waitaof,
        )
        if (
            self.cache
            and isinstance(self.cache, SupportsClientTracking)
            and connection.tracking_client_id != self.cache.get_client_id(connection)
        ):
            self.cache.reset()
            await connection.update_tracking_client(
                True, self.cache.get_client_id(connection)
            )
        try:
            if self.cache and command not in READONLY_COMMANDS:
                self.cache.invalidate(*KeySpec.extract_keys(command, *args))
            request = await connection.create_request(
                command,
                *args,
                noreply=self.noreply,
                decode=options.get("decode", self._decodecontext.get()),
                encoding=self._encodingcontext.get(),
            )
            maybe_wait = [
                await self._ensure_wait(command, connection),
                await self._ensure_persistence(command, connection),
            ]
            reply = await request
            await asyncio.gather(*maybe_wait)
            if self.noreply:
                return None  # type: ignore
            if isinstance(callback, AsyncPreProcessingCallback):
                await callback.pre_process(
                    self, reply, version=self.protocol_version, **options
                )
            return callback(
                reply,
                version=self.protocol_version,
                **options,
            )
        except RedisError:
            connection.disconnect()
            raise
        finally:
            self._ensure_server_version(connection.server_version)
            if not quick_release or self.requires_wait or self.requires_waitaof:
                pool.release(connection)

    @overload
    def decoding(
        self, mode: Literal[False], encoding: Optional[str] = None
    ) -> ContextManager[Redis[bytes]]: ...

    @overload
    def decoding(
        self, mode: Literal[True], encoding: Optional[str] = None
    ) -> ContextManager[Redis[str]]: ...

    @contextlib.contextmanager
    @versionadded(version="4.8.0")
    def decoding(
        self, mode: bool, encoding: Optional[str] = None
    ) -> Iterator[Redis[Any]]:
        """
        Context manager to temporarily change the decoding behavior
        of the client

        :param mode: Whether to decode or not
        :param encoding: Optional encoding to use if decoding. If not provided
         the :paramref:`~coredis.Redis.encoding` parameter provided to the client will
         be used.

        Example::

            client = coredis.Redis(decode_responses=True)
            await client.set("fubar", "baz")
            assert await client.get("fubar") == "baz"
            with client.decoding(False):
                assert await client.get("fubar") == b"baz"
                with client.decoding(True):
                    assert await client.get("fubar") == "baz"

        """
        prev_decode = self._decodecontext.get()
        prev_encoding = self._encodingcontext.get()
        self._decodecontext.set(mode)
        self._encodingcontext.set(encoding)
        try:
            yield self
        finally:
            self._decodecontext.set(prev_decode)
            self._encodingcontext.set(prev_encoding)

    def monitor(self) -> Monitor[AnyStr]:
        """
        Return an instance of a :class:`~coredis.commands.monitor.Monitor`

        The monitor can be used as an async iterator or individual commands
        can be fetched via :meth:`~coredis.commands.monitor.Monitor.get_command`.
        """
        return Monitor[AnyStr](self)

    def pubsub(
        self,
        ignore_subscribe_messages: bool = False,
        retry_policy: Optional[RetryPolicy] = None,
        **kwargs: Any,
    ) -> PubSub[AnyStr]:
        """
        Return a Pub/Sub instance that can be used to subscribe to channels
        and patterns and receive messages that get published to them.

        :param ignore_subscribe_messages: Whether to skip subscription
         acknowledgement messages
        :param retry_policy: An explicit retry policy to use in the subscriber.
        """

        return PubSub[AnyStr](
            self.connection_pool,
            ignore_subscribe_messages=ignore_subscribe_messages,
            retry_policy=retry_policy,
            **kwargs,
        )

    async def pipeline(
        self,
        transaction: Optional[bool] = True,
        watches: Optional[Parameters[KeyT]] = None,
        timeout: Optional[float] = None,
    ) -> "coredis.pipeline.Pipeline[AnyStr]":
        """
        Returns a new pipeline object that can queue multiple commands for
        batch execution.

        :param transaction: indicates whether all commands should be executed atomically.
        :param watches: If :paramref:`transaction` is True these keys are watched for external
         changes during the transaction.
        :param timeout: If specified this value will take precedence over
         :paramref:`Redis.stream_timeout`
        """
        from coredis.pipeline import Pipeline

        return Pipeline[AnyStr].proxy(self, transaction, watches, timeout)

    async def transaction(
        self,
        func: Callable[["coredis.pipeline.Pipeline[AnyStr]"], Coroutine[Any, Any, Any]],
        *watches: KeyT,
        value_from_callable: bool = False,
        watch_delay: Optional[float] = None,
        **kwargs: Any,
    ) -> Optional[Any]:
        """
        Convenience method for executing the callable :paramref:`func` as a
        transaction while watching all keys specified in :paramref:`watches`.

        :param func: callable should expect a single argument which is a
         :class:`coredis.pipeline.Pipeline` object retrieved by calling
         :meth:`~coredis.Redis.pipeline`.
        :param watches: The keys to watch during the transaction
        :param value_from_callable: Whether to return the result of transaction or the value
         returned from :paramref:`func`
        """
        async with await self.pipeline(True) as pipe:
            while True:
                try:
                    if watches:
                        await pipe.watch(*watches)
                    func_value = await func(pipe)
                    exec_value = await pipe.execute()
                    return func_value if value_from_callable else exec_value
                except WatchError:
                    if watch_delay is not None and watch_delay > 0:
                        await asyncio.sleep(watch_delay)
                    continue
